##
##

import logging
import io
import subprocess
from typing import Union, List

logger = logging.getLogger('couchformation.exec.process')
logger.addHandler(logging.NullHandler())


def cmd_exec(command: Union[str, List[str]], directory: Union[str, None] = None) -> io.BytesIO:
    buffer = io.BytesIO()

    p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=directory)

    while True:
        data = p.stdout.read()
        if not data:
            break
        buffer.write(data)

    p.communicate()

    if p.returncode != 0:
        raise ValueError("command exited with non-zero return code")

    buffer.seek(0)
    return buffer


def get_output_buffer(buffer: io.BytesIO):
    while True:
        line = buffer.readline()
        if not line:
            break
        line_string = line.decode("utf-8")
        yield line_string



##
##

from typing import Union, List
import subprocess
import logging
import io

logger = logging.getLogger('couchformation.provisioner.shell')
logger.addHandler(logging.NullHandler())


class ShellCommandError(Exception):
    pass


class RCNotZero(Exception):
    pass


class RunShellCommand(object):

    def __init__(self):
        pass

    @staticmethod
    def cmd_exec(command: Union[str, List[str]], directory: str):
        buffer = io.BytesIO()
        logger.debug(f"Shell command: {' '.join(command)}")

        p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=directory)

        while True:
            data = p.stdout.read()
            if not data:
                break
            buffer.write(data)

        p.communicate()
        buffer.seek(0)

        if p.returncode != 0:
            output_text = buffer.read().decode("utf-8").strip()
            raise ShellCommandError(f"command error: {output_text}")

        return buffer

    @staticmethod
    def cmd_output(command: Union[str, List[str]], directory: str, split: bool = False, split_sep: str = None, no_raise: bool = False):
        out_lines = []
        try:
            output: io.BytesIO = RunShellCommand().cmd_exec(command, directory)
        except ShellCommandError as err:
            if not no_raise:
                raise RCNotZero(err)
            else:
                return None

        while True:
            line = output.readline()
            if not line:
                break
            line_string = line.decode("utf-8").strip()
            if len(line_string) > 0:
                if split:
                    items = line_string.split(split_sep)
                    out_lines.append(items)
                else:
                    out_lines.append(line_string)

        return out_lines



##
##

import base64
import ipaddress
import datetime
from typing import List
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import load_pem_private_key, load_der_private_key
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID


class CertMgr(object):

    def __init__(self):
        pass

    @staticmethod
    def private_key(size=2048):
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=size,
            backend=default_backend(),
        )
        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        ).decode()
        return base64.b64encode(pem.encode()).decode()

    @staticmethod
    def certificate_hostname(filename: str, private_key: str, domain_name: str = None, alt_name: List[str] = None, alt_ip_list: List[str] = None):
        one_day = datetime.timedelta(1, 0, 0)
        with open(private_key, 'r') as pem_in:
            privkey = pem_in.read()
        private_key = load_pem_private_key(privkey.encode(), None, default_backend())
        public_key = private_key.public_key()

        cert_name = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "California"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, "Santa Clara"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Couchbase")
        ])

        builder = x509.CertificateBuilder()
        builder = builder.subject_name(cert_name)
        builder = builder.issuer_name(cert_name)
        builder = builder.not_valid_before(datetime.datetime.today() - one_day)
        builder = builder.not_valid_after(datetime.datetime.today() + (one_day * 365 * 10))
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.public_key(public_key)

        host_names = []

        if alt_name is not None:
            for name in alt_name:
                host_names.append(x509.DNSName(name))

        if alt_ip_list is not None:
            for ip in alt_ip_list:
                host_names.append(x509.IPAddress(ipaddress.ip_address(ip)))

        if domain_name is not None:
            host_names.append(x509.DNSName(f"*.{domain_name}"))

        host_names.append(x509.IPAddress(ipaddress.ip_address('127.0.0.1')))

        builder = builder.add_extension(x509.SubjectAlternativeName(host_names), critical=True)
        builder = builder.add_extension(x509.BasicConstraints(ca=False, path_length=None), critical=True)
        builder = builder.add_extension(x509.SubjectKeyIdentifier.from_public_key(public_key), critical=True)
        builder = builder.add_extension(x509.AuthorityKeyIdentifier.from_issuer_public_key(public_key), critical=True)
        builder = builder.add_extension(x509.ExtendedKeyUsage([ExtendedKeyUsageOID.SERVER_AUTH]), critical=True)
        builder = builder.add_extension(x509.KeyUsage(
            digital_signature=True,
            key_encipherment=True,
            data_encipherment=False,
            content_commitment=False,
            key_agreement=False,
            key_cert_sign=False,
            crl_sign=False,
            encipher_only=False,
            decipher_only=False), critical=True)

        certificate = builder.sign(private_key=private_key, algorithm=hashes.SHA256(), backend=default_backend())

        cert = certificate.public_bytes(serialization.Encoding.PEM).decode('utf-8')
        with open(filename, 'w') as f:
            f.write(cert)

    @staticmethod
    def certificate_basic(filename: str, private_key: str):
        one_day = datetime.timedelta(1, 0, 0)
        with open(private_key, 'r') as pem_in:
            privkey = pem_in.read()
        private_key = load_pem_private_key(privkey.encode(), None, default_backend())
        public_key = private_key.public_key()

        cert_name = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "California"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, "Santa Clara"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Couchbase")
        ])

        builder = x509.CertificateBuilder()
        builder = builder.subject_name(cert_name)
        builder = builder.issuer_name(cert_name)
        builder = builder.not_valid_before(datetime.datetime.today() - one_day)
        builder = builder.not_valid_after(datetime.datetime.today() + (one_day * 365 * 10))
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.public_key(public_key)

        certificate = builder.sign(private_key=private_key, algorithm=hashes.SHA256(), backend=default_backend())

        cert = certificate.public_bytes(serialization.Encoding.PEM).decode('utf-8')
        with open(filename, 'w') as f:
            f.write(cert)

    @staticmethod
    def certificate_ca(key: str):
        one_day = datetime.timedelta(1, 0, 0)
        private_key = load_pem_private_key(base64.b64decode(key), None, default_backend())
        public_key = private_key.public_key()

        cert_name = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "California"),
            x509.NameAttribute(NameOID.LOCALITY_NAME, "Santa Clara"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Couchbase")
        ])

        builder = x509.CertificateBuilder()
        builder = builder.subject_name(cert_name)
        builder = builder.issuer_name(cert_name)
        builder = builder.not_valid_before(datetime.datetime.today() - one_day)
        builder = builder.not_valid_after(datetime.datetime.today() + (one_day * 365 * 10))
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.public_key(public_key)
        builder = builder.add_extension(
            x509.BasicConstraints(ca=True, path_length=None), critical=True,
        )
        certificate = builder.sign(
            private_key=private_key, algorithm=hashes.SHA256(),
            backend=default_backend()
        )

        key_bytes = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        )

        cert_bytes = certificate.public_bytes(
            encoding=serialization.Encoding.PEM,
        )

        return base64.b64encode(cert_bytes).decode()



##
##

import logging
from couchformation.gcp.driver.base import CloudBase, GCPDriverError, EmptyResultSet
from couchformation.gcp.driver.constants import ComputeTypes
import couchformation.constants as C

logger = logging.getLogger('couchformation.gcp.driver.machine')
logger.addHandler(logging.NullHandler())
logging.getLogger("googleapiclient").setLevel(logging.ERROR)


class MachineType(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, zone: str, architecture: str = 'x86_64') -> list:
        machine_type_list = []
        if architecture == 'arm64':
            filter_string = "cpuPlatform = \"Ampere Altra\""
        else:
            filter_string = None

        try:
            request = self.gcp_client.machineTypes().list(project=self.gcp_project, zone=zone, filter=filter_string)
            while request is not None:
                response = request.execute()
                for machine_type in response['items']:
                    if not machine_type['name'].startswith(tuple(ComputeTypes().as_list())):
                        continue
                    config_block = {'name': machine_type['name'],
                                    'id': machine_type['id'],
                                    'cpu': int(machine_type['guestCpus']),
                                    'memory': int(machine_type['memoryMb']),
                                    'description': machine_type['description']}
                    machine_type_list.append(config_block)
                request = self.gcp_client.machineTypes().list_next(previous_request=request, previous_response=response)
        except Exception as err:
            raise GCPDriverError(f"error listing machine types: {err}")

        if len(machine_type_list) == 0:
            raise EmptyResultSet(f"no instance types found")

        return machine_type_list

    def get_machine_types(self, zone: str, architecture: str = 'x86_64'):
        result_list = []
        machine_list = self.list(zone, architecture)
        machine_list = sorted(machine_list, key=lambda m: m['name'])

        for machine_type in C.MACHINE_TYPES:
            machine = next((m for m in machine_list if m['cpu'] == machine_type['cpu'] and m['memory'] == machine_type['memory']), None)
            if not machine:
                continue
            machine.update(dict(machine_type=machine_type['name']))
            result_list.append(machine)

        return result_list

    def get_machine(self, name: str, zone: str, architecture: str = 'x86_64'):
        machine_list = self.get_machine_types(zone, architecture)
        return next((m for m in machine_list if m['machine_type'] == name), None)

    def details(self, machine_type: str) -> dict:
        try:
            request = self.gcp_client.machineTypes().get(project=self.gcp_project, zone=self.gcp_zone, machineType=machine_type)
            response = request.execute()
            return {'name': response['name'],
                    'id': response['id'],
                    'cpu': int(response['guestCpus']),
                    'memory': int(response['memoryMb']),
                    'description': response['description']}
        except Exception as err:
            GCPDriverError(f"error getting machine type details: {err}")



##
##

import logging
from typing import List, Union
import googleapiclient.errors
from couchformation.gcp.driver.base import CloudBase, GCPDriverError, EmptyResultSet

logger = logging.getLogger('couchformation.gcp.driver.disk')
logger.addHandler(logging.NullHandler())
logging.getLogger("googleapiclient").setLevel(logging.ERROR)


class Disk(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, zone: str) -> List[dict]:
        disk_list = []

        try:
            request = self.gcp_client.disks().list(project=self.gcp_project, zone=zone)
            while request is not None:
                response = request.execute()

                for disk in response['items']:
                    disk_list.append(disk)
                request = self.gcp_client.disks().list_next(previous_request=request, previous_response=response)
        except Exception as err:
            raise GCPDriverError(f"error listing disks: {err}")

        if len(disk_list) == 0:
            raise EmptyResultSet(f"no disks found")
        else:
            return disk_list

    def create(self, name: str, zone: str, size: str, disk_type: str = "pd-ssd") -> str:
        operation = {}
        disk_body = {
            "sizeGb": str(round(float(size))),
            "name": name,
            "type": f"zones/{zone}/diskTypes/{disk_type}"
        }
        try:
            request = self.gcp_client.disks().insert(project=self.gcp_project, zone=zone, body=disk_body)
            operation = request.execute()
            self.wait_for_zone_operation(operation['name'], zone)
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "alreadyExists":
                raise GCPDriverError(f"can not create disk: {err}")
        except Exception as err:
            raise GCPDriverError(f"error creating disk: {err}")

        return operation.get('targetLink')

    def delete(self, disk: str, zone: str) -> None:
        try:
            request = self.gcp_client.disks().delete(project=self.gcp_project, zone=zone, disk=disk)
            operation = request.execute()
            self.wait_for_zone_operation(operation['name'], zone)
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not delete disk: {err}")
        except Exception as err:
            raise GCPDriverError(f"error deleting disk: {err}")

    def details(self, disk: str, zone: str) -> Union[dict, None]:
        try:
            request = self.gcp_client.disks().get(project=self.gcp_project, zone=zone, disk=disk)
            result = request.execute()
            return result
        except googleapiclient.errors.HttpError as err:
            error_details = err.error_details[0].get('reason')
            if error_details != "notFound":
                raise GCPDriverError(f"can not find disk: {err}")
            return None
        except Exception as err:
            raise GCPDriverError(f"error getting disk: {err}")

    def find(self, disk: str) -> Union[dict, None]:
        for zone in self.gcp_zone_list:
            result = self.details(disk, zone)
            if result:
                return result
        return None



##
##

import time
import logging
import os
import configparser
from functools import wraps
from typing import Union, List, Callable
from azure.identity import AzureCliCredential
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.dns import DnsManagementClient
from azure.mgmt.privatedns import PrivateDnsManagementClient
from azure.mgmt.resource.resources import ResourceManagementClient
from azure.mgmt.resource.subscriptions import SubscriptionClient
from couchformation.config import AuthMode
from couchformation.exception import FatalError, NonFatalError
from couchformation.azure.driver.constants import get_auth_directory, get_config_default, get_config_main
from couchformation.azure.driver.constants import AzureDiskTiers
from couchformation.exec.process import cmd_exec

logger = logging.getLogger('couchformation.azure.driver.base')
logger.addHandler(logging.NullHandler())
logging.getLogger("azure").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


def auth_retry(retry_count=10,
               factor=0.01
               ) -> Callable:
    def retry_handler(func):
        @wraps(func)
        def f_wrapper(*args, **kwargs):
            for retry_number in range(retry_count + 1):
                try:
                    return func(*args, **kwargs)
                except Exception as err:
                    if retry_number == retry_count:
                        logger.debug(f"{func.__name__} retry limit exceeded: {err}")
                        raise
                    login_cmd = ['az', 'login']
                    cmd_exec(login_cmd)
                    wait = factor
                    wait *= (2 ** (retry_number + 1))
                    time.sleep(wait)
        return f_wrapper
    return retry_handler


class AzureDriverError(FatalError):
    pass


class EmptyResultSet(NonFatalError):
    pass


class CloudBase(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.auth_directory = get_auth_directory()
        self.config_default = get_config_default()
        self.config_main = get_config_main()
        self.cloud_name = 'AzureCloud'
        self.local_context = None
        self.azure_subscription_id = None
        self.credential = None
        self.azure_resource_group = None
        self.azure_location = None
        self.azure_availability_zones = []
        self.azure_zone = None
        self.subscription_client = None

        if not parameters.get('auth_mode') or AuthMode[parameters.get('auth_mode')] == AuthMode.default:
            self.credential, self.azure_subscription_id, self.azure_tenant_id = self.default_auth()
        else:
            raise AzureDriverError(f"Unsupported auth mode {parameters.get('auth_mode')}")

        if not self.credential or not self.azure_subscription_id:
            raise AzureDriverError("unauthorized (use az login)")

        self.resource_client = ResourceManagementClient(self.credential, self.azure_subscription_id)
        self.compute_client = ComputeManagementClient(self.credential, self.azure_subscription_id)
        self.network_client = NetworkManagementClient(self.credential, self.azure_subscription_id)
        self.dns_client = DnsManagementClient(self.credential, self.azure_subscription_id)
        self.private_dns_client = PrivateDnsManagementClient(self.credential, self.azure_subscription_id)

        self.azure_location = parameters.get('region')

        self.zones()

    @auth_retry()
    def default_auth(self):
        try:
            credential = AzureCliCredential()
            subscription_client = SubscriptionClient(credential)
            subscriptions = subscription_client.subscriptions.list()
            azure_subscription_id = next((s.subscription_id for s in subscriptions), None)
            azure_tenant_id = credential.tenant_id
            return credential, azure_subscription_id, azure_tenant_id
        except Exception as err:
            raise AzureDriverError(f"Azure: unauthorized (use az login): {err}")

    def test_session(self):
        if len(self.azure_availability_zones) == 0:
            raise AzureDriverError(f"Unable to determine availability zones for location {self.azure_location}")

    @property
    def subscription_id(self):
        return self.azure_subscription_id

    @property
    def tenant_id(self):
        return self.azure_tenant_id

    @staticmethod
    def disk_size_to_tier(value: Union[int, str]):
        size = int(value)
        size_list = [int(i['disk_size']) for i in AzureDiskTiers.disk_tier_list]
        value = min([s for s in size_list if s >= size])
        return next(t for t in AzureDiskTiers.disk_tier_list if t['disk_size'] == str(value))

    @staticmethod
    def disk_caching(value: Union[int, str], ultra: bool = False):
        size = int(value)
        if size > 4095 or ultra:
            return "None"
        else:
            return "ReadWrite"

    def read_config(self):
        if os.path.exists(self.config_main):
            config = configparser.ConfigParser()
            try:
                config.read(self.config_main)
            except Exception as err:
                raise AzureDriverError(f"can not read config file {self.config_main}: {err}")

            if 'cloud' in config:
                if 'name' in config['cloud']:
                    self.cloud_name = config['cloud']['name']

            if 'local_context' in config:
                try:
                    self.local_context = list(config['local_context'].keys())[0]
                except IndexError:
                    pass

        if os.path.exists(self.config_default):
            config = configparser.ConfigParser()
            try:
                config.read(self.config_default)
            except Exception as err:
                raise AzureDriverError(f"can not read config file {self.config_default}: {err}")

            if self.cloud_name in config:
                self.azure_subscription_id = config[self.cloud_name].get('subscription', None)

    def zones(self) -> list:
        zone_list = self.compute_client.resource_skus.list(filter=f"location eq '{self.azure_location}'")
        for group in list(zone_list):
            if group.resource_type == 'virtualMachines':
                for resource_location in group.location_info:
                    for zone_number in resource_location.zones:
                        self.azure_availability_zones.append(zone_number)

        self.azure_availability_zones = sorted(set(self.azure_availability_zones))

        if len(self.azure_availability_zones) == 0:
            raise AzureDriverError("can not get Azure availability zones")

        self.azure_zone = self.azure_availability_zones[0]
        return self.azure_availability_zones

    def create_rg(self, name: str, location: str, tags: Union[dict, None] = None) -> dict:
        if not tags:
            tags = {}
        if not tags.get('type'):
            tags.update({"type": "couch-formation"})
        try:
            if self.resource_client.resource_groups.check_existence(name):
                return self.get_rg(name, location)
            else:
                result = self.resource_client.resource_groups.create_or_update(
                    name,
                    {
                        "location": location,
                        "tags": tags
                    }
                )
                return result.__dict__
        except Exception as err:
            raise AzureDriverError(f"error creating resource group: {err}")

    def get_rg(self, name: str, location: str) -> Union[dict, None]:
        try:
            if self.resource_client.resource_groups.check_existence(name):
                result = self.resource_client.resource_groups.get(name)
                if result.location == location:
                    return result.__dict__
        except Exception as err:
            raise AzureDriverError(f"error getting resource group: {err}")

        return None

    def list_rg(self, location: Union[str, None] = None, filter_keys_exist: Union[List[str], None] = None) -> List[dict]:
        rg_list = []

        try:
            resource_groups = self.resource_client.resource_groups.list()
        except Exception as err:
            raise AzureDriverError(f"error getting resource groups: {err}")

        for group in list(resource_groups):
            if location:
                if group.location != location:
                    continue
            rg_block = {'location': group.location,
                        'name': group.name,
                        'id': group.id}
            rg_block.update(self.process_tags(group.tags))
            if filter_keys_exist:
                if not all(key in rg_block for key in filter_keys_exist):
                    continue
            rg_list.append(rg_block)

        if len(rg_list) == 0:
            raise EmptyResultSet(f"no resource groups found")

        return rg_list

    def delete_rg(self, name: str):
        try:
            if self.resource_client.resource_groups.check_existence(name):
                request = self.resource_client.resource_groups.begin_delete(name)
                request.wait()
        except Exception as err:
            raise AzureDriverError(f"error deleting resource group: {err}")

    def list_locations(self) -> List[dict]:
        location_list = []
        locations = self.subscription_client.subscriptions.list_locations(self.azure_subscription_id)
        for group in list(locations):
            location_block = {
                'name': group.name,
                'display_name': group.display_name
            }
            location_list.append(location_block)
        return location_list

    @staticmethod
    def process_tags(struct: dict) -> dict:
        block = {}
        if struct:
            for tag in struct:
                block.update({tag.lower() + '_tag': struct[tag]})
        block = dict(sorted(block.items()))
        return block

    def rg_switch(self):
        image_rg = f"cf-image-{self.azure_location}-rg"
        if self.get_rg(image_rg, self.azure_location):
            resource_group = image_rg
        else:
            resource_group = self.azure_resource_group
        return resource_group

    @property
    def region(self):
        return self.azure_location



##
##

import paramiko
import paramiko.util
import subprocess
import logging
import io
import time
import socket
import os
import couchformation.constants as C

logger = logging.getLogger('couchformation.provisioner.ssh')
logger.addHandler(logging.NullHandler())


class SSHError(Exception):
    pass


old_factory = logging.getLogRecordFactory()


class KeepAliveFilter(logging.Filter):
    def filter(self, record):
        return record.msg.find('keepalive@lag.net') < 0


def record_factory_factory(context_id):
    def record_factory(*args, **kwargs):
        record = old_factory(*args, **kwargs)
        record.ip_address = context_id
        return record
    return record_factory


class CustomLogFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"{C.FORMAT_TIMESTAMP} (%(ip_address)s) [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.INFO: f"{C.FORMAT_TIMESTAMP} (%(ip_address)s) [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.WARNING: f"{C.FORMAT_TIMESTAMP} (%(ip_address)s) [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.ERROR: f"{C.FORMAT_TIMESTAMP} (%(ip_address)s) [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"{C.FORMAT_TIMESTAMP} (%(ip_address)s) [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        if logging.DEBUG >= logging.root.level:
            log_fmt += C.FORMAT_EXTRA
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class RunSSHCommand(object):

    def __init__(self, ssh_key: str, ssh_user: str, hostname: str, command: str, working_dir: str):
        self.ssh_key = ssh_key
        self.ssh_user = ssh_user
        self.hostname = hostname
        self.command = command

        self.file_output = logging.getLogger("paramiko")
        self.file_output.propagate = False
        log_file = os.path.join(working_dir, 'connect.log')
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(CustomLogFormatter())
        self.file_output.addHandler(file_handler)
        # file_output.setLevel(logging.root.level)
        self.file_output.setLevel(logging.DEBUG)
        logging.setLogRecordFactory(record_factory_factory(hostname))

        paramiko.util.get_logger('paramiko.transport').addFilter(KeepAliveFilter())

    def ssh_connect(self, retry_count=35, factor=0.5):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.WarningPolicy())

        for retry_number in range(retry_count):
            try:
                ssh.connect(self.hostname, username=self.ssh_user, key_filename=self.ssh_key, timeout=10, auth_timeout=10, banner_timeout=10, allow_agent=False)
                ssh.get_transport().set_keepalive(5)
                return ssh
            except paramiko.ssh_exception.BadHostKeyException as err:
                raise RuntimeError(f"host key mismatch for {self.hostname}: {err}")
            except paramiko.ssh_exception.AuthenticationException as err:
                raise RuntimeError(f"failed to authenticate to {self.hostname}: {err}")
            except (paramiko.ssh_exception.SSHException, TimeoutError, socket.timeout) as err:
                n_retry = retry_number + 1
                if n_retry == retry_count:
                    raise RuntimeError(f"can not connect to {self.hostname}: {err}")
                logger.info(f"Waiting for an SSH connection to {self.hostname}")
                self.file_output.error(f"retrying SSH connect: count {n_retry}: {err}")
                wait = factor
                wait *= n_retry
                time.sleep(wait)

    def exec(self, connect_retry=35, command_retry=10, factor=0.5):
        bufsize = 4096

        ssh = self.ssh_connect(retry_count=connect_retry, factor=factor)

        for retry_number in range(command_retry):
            try:
                stdin, stdout, stderr = ssh.exec_command(self.command, bufsize=bufsize, timeout=10)
                channel = stdout.channel
                stdin.close()
                channel.shutdown_write()
                exit_code = channel.recv_exit_status()
                timeout = 5
                end_time = time.time() + timeout
                while not stdout.channel.eof_received:
                    time.sleep(0.5)
                    if time.time() > end_time:
                        stdout.channel.close()
                        break
                ssh.close()
                return exit_code, stdout, stderr
            except Exception as err:
                n_retry = retry_number + 1
                if n_retry == command_retry:
                    raise RuntimeError(f"command failed on {self.hostname}: {err}")
                logger.info(f"Retrying command on {self.hostname}")
                self.file_output.error(f"retrying command exec: count {n_retry}: {err}")
                wait = factor
                wait *= n_retry
                time.sleep(wait)


class RunSSHCommandExec(object):

    def __init__(self):
        pass

    @staticmethod
    def exec(ssh_key: str, ssh_user: str, hostname: str, command: str, directory: str):
        buffer = io.BytesIO()
        logger.debug(f"Shell command: {command}")

        ssh_cmd = f"""ssh -i {ssh_key} -l {ssh_user} {hostname} '{command}'"""

        p = subprocess.Popen(ssh_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=directory)

        while True:
            data = p.stdout.read()
            if not data:
                break
            buffer.write(data)

        p.communicate()
        buffer.seek(0)

        return p.returncode, buffer



##
##

import attr
from attr.validators import instance_of as io
from typing import Iterable
from enum import Enum

CLOUD_KEY = "aws"

aws_storage_matrix = {
    99: "3000",
    199: "5000",
    299: "6000",
    399: "8000",
    499: "9000",
    599: "10000",
    699: "12000",
    799: "13000",
    899: "14000",
    999: "16000",
    16384: "16000"
}


aws_arch_matrix = {
    'x86_64': 'zone',
    'arm64': 'zone',
    'arm64_mac': 'host'
}


class PlacementType(Enum):
    ZONE = "zone"
    HOST = "host"


@attr.s
class AWSTag(object):
    Key = attr.ib(validator=io(str))
    Value = attr.ib(validator=io(str))

    @property
    def as_dict(self):
        return self.__dict__


@attr.s
class AWSTagStruct(object):
    ResourceType = attr.ib(validator=io(str))
    Tags = attr.ib(validator=io(Iterable))

    @classmethod
    def build(cls, resource: str):
        return cls(
            resource,
            []
        )

    def add(self, obj: AWSTag):
        self.Tags.append(obj.as_dict)
        return self

    @property
    def as_dict(self):
        return self.__dict__


@attr.s
class EbsVolume(object):
    VolumeType = attr.ib(validator=io(str))
    VolumeSize = attr.ib(validator=io(int))
    Iops = attr.ib(validator=io(int))

    @classmethod
    def build(cls, vol_type: str, vol_size: int, vol_iops: int):
        return cls(
            vol_type,
            vol_size,
            vol_iops
        )

    @property
    def as_dict(self):
        return self.__dict__


@attr.s
class AWSEbsDisk(object):
    DeviceName = attr.ib(validator=io(str))
    Ebs = attr.ib(validator=io(dict))

    @classmethod
    def build(cls, device: str, obj: EbsVolume):
        return cls(
            device,
            obj.as_dict
        )

    @property
    def as_dict(self):
        return self.__dict__


@attr.s
class AWSEbsDiskTypes(object):
    ebs_type_list = [
        {
            "type": "standard",
            "iops": None,
            "max": None
        },
        {
            "type": "io1",
            "iops": 3000,
            "max": 64000
        },
        {
            "type": "io2",
            "iops": 3000,
            "max": 64000
        },
        {
            "type": "gp2",
            "iops": None,
            "max": None
        },
        {
            "type": "sc1",
            "iops": None,
            "max": None
        },
        {
            "type": "st1",
            "iops": None,
            "max": None
        },
        {
            "type": "gp3",
            "iops": 3000,
            "max": 16000
        }
    ]


@attr.s
class AWSImageOwners(object):
    image_owner_list = [
        {
            "owner_id": "099720109477",
            "description": "Ubuntu Linux",
            "os_id": "ubuntu",
            "user": "ubuntu",
            "feature": None,
            "pattern": r"ubuntu/images/hvm-ssd/ubuntu-*-server-*",
            "version": r"ubuntu/images/hvm-ssd/ubuntu-.*-(.+?)-.*-server-.*"
        },
        {
            "owner_id": "099720109477",
            "description": "Ubuntu Linux",
            "os_id": "ubuntu-pro",
            "user": "ubuntu",
            "feature": None,
            "pattern": r"ubuntu-pro-server/images/hvm-ssd/ubuntu-*-pro-server-*",
            "version": r"ubuntu-pro-server/images/hvm-ssd/ubuntu-.*-(.+?)-.*-pro-server-.*"
        },
        {
            "owner_id": "125523088429",
            "description": "CentOS Linux",
            "os_id": "centos",
            "user": "centos",
            "feature": None,
            "pattern": r"CentOS Stream * * *",
            "version": r"CentOS Stream (.+?) .* .*",
        },
        {
            "owner_id": "309956199498",
            "description": "RedHat Linux",
            "os_id": "rhel",
            "user": "ec2-user",
            "feature": None,
            "pattern": r"RHEL-?.?.?_HVM-*-*-*",
            "version": r"RHEL-(.+?).[0-9].[0-9]_HVM-[0-9]*-.*-.*"
        },
        {
            "owner_id": "013907871322",
            "description": "Suse Linux",
            "os_id": "sles",
            "user": "ec2-user",
            "feature": None,
            "pattern": r"suse-sles-*-sp?-v*-hvm-ssd-*",
            "version": r"suse-sles-(.+?)-sp[0-9]-v[0-9]*-hvm-ssd-.*"
        },
        {
            "owner_id": "679593333241",
            "description": "openSUSE Leap",
            "os_id": "opensuse-leap",
            "user": "ec2-user",
            "feature": None,
            "pattern": r"openSUSE-Leap-*-?-v*-hvm-ssd-*-*",
            "version": r"openSUSE-Leap-(.+?)-[0-9]-v[0-9]*-hvm-ssd-.*-.*"
        },
        {
            "owner_id": "379101102735",
            "description": "Debian 9 and earlier",
            "os_id": "debian",
            "user": "admin",
            "feature": None,
            "pattern": r"debian-*-*-*-*",
            "version": r"debian-(.+?)-.*-[0-9]*-[0-9]*"
        },
        {
            "owner_id": "136693071363",
            "description": "Debian 10 and later",
            "os_id": "debian",
            "user": "admin",
            "feature": None,
            "pattern": r"debian-*-*-*-*",
            "version": r"debian-(.+?)-.*-[0-9]*-[0-9]*"
        },
        {
            "owner_id": "131827586825",
            "description": "Oracle Linux",
            "os_id": "ol",
            "user": "ec2-user",
            "feature": None,
            "pattern": r"OL?.?-*-HVM-*-*-*",
            "version": r"OL(.+?).[0-9]-.*-HVM-[0-9]*-[0-9]*-[0-9]*"
        },
        {
            "owner_id": "125523088429",
            "description": "Fedora CoreOS Linux",
            "os_id": "fedora",
            "user": "core",
            "feature": None,
            "pattern": r"fedora-coreos-*.*.?.?-*",
            "version": r"fedora-coreos-(.+?).[0-9]*.[0-9].[0-9]-.*"
        },
        {
            "owner_id": "137112412989",
            "description": "Amazon Linux",
            "os_id": "amzn",
            "user": "ec2-user",
            "feature": None,
            "pattern": r"*-ami-*.*.0-*",
            "version": r"[a-z]*(.+?)-ami-.*.[0-9]*.0-.*"
        },
        {
            "owner_id": "647457786197",
            "description": "Arch Linux",
            "os_id": "arch",
            "user": "arch",
            "feature": None,
            "pattern": r"arch-linux-*-hvm-*.*.*.*-*",
            "version": r"none"
        },
        {
            "owner_id": "792107900819",
            "description": "Rocky Linux",
            "os_id": "rocky",
            "user": "rocky",
            "feature": None,
            "pattern": r"Rocky-?-EC2-?.?-*.0*",
            "version": r"Rocky-(.+?)-EC2-[0-9].[0-9]-[0-9]*.0.*"
        },
        {
            "owner_id": "801119661308",
            "description": "Windows",
            "os_id": "windows",
            "user": "Administrator",
            "feature": None,
            "pattern": r"Windows_Server-*-English-Full-Base-*.*.*",
            "version": r"Windows_Server-(.+?)-English-Full-Base-[0-9]*.[0-9]*.[0-9]*"
        },
        {
            "owner_id": "801119661308",
            "description": "Windows",
            "os_id": "windows",
            "user": "Administrator",
            "feature": "vmp",
            "pattern": r"Windows_Server-*-English-Full-HyperV-*.*.*",
            "version": r"Windows_Server-(.+?)-English-Full-HyperV-[0-9]*.[0-9]*.[0-9]*"
        },
        {
            "owner_id": "634519214787",
            "description": "macOS",
            "os_id": "macos",
            "user": "ec2-user",
            "feature": None,
            "pattern": r"amzn-ec2-macos-*.*.*-*-*-arm64",
            "version": r"amzn-ec2-macos-(.+?).[0-9]*.[0-9]*-[0-9]*-[0-9]*-arm64"
        },
    ]


@attr.s
class ComputeTypes(object):
    general_purpose = ['m5', 'm5a', 'm7g', "mac2", "mac2-m2", "mac2-m2pro"]
    compute_optimized = ['c5', 'c5a', 'c7g']
    memory_optimized = ['r5', 'r5a', 'r7g']

    def as_list(self) -> list:
        flat_list = []
        for element in [self.general_purpose, self.compute_optimized, self.memory_optimized]:
            flat_list.extend(element)
        return flat_list


@attr.s
class ArchitectureTypes(object):
    list = ['x86_64', 'arm64']



##
##

import attr


@attr.s
class ContainerBuildMap(object):
    build_map = {
        'cbs': 'couchbase/server',
        'sgw': 'couchbase/sync-gateway'
    }

    def image(self, build):
        return self.build_map.get(build)



##
##

import os
import logging
import jinja2
from couchformation.config import get_state_dir
from couchformation.docker.driver.container import Container
import couchformation.constants as C

logger = logging.getLogger('couchformation.provisioner.docker')
logger.addHandler(logging.NullHandler())
logging.getLogger("docker").setLevel(logging.WARNING)


class CustomLogFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.INFO: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.WARNING: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.ERROR: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}",
        logging.CRITICAL: f"{C.FORMAT_TIMESTAMP} [{C.FORMAT_LEVEL}] {C.FORMAT_MESSAGE}"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        if logging.DEBUG >= logging.root.level:
            log_fmt += C.FORMAT_EXTRA
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class ContainerExec(object):

    def __init__(self, parameters: dict, command: str = '', root: bool = True):
        self.parameters = parameters
        self.command = command
        self.root = root
        self.container_name = self.parameters.get('name')
        self.service = self.parameters.get('service')
        self.project = self.parameters.get('project')
        self.public_ip = self.parameters.get('public_ip')
        self.private_ip = self.parameters.get('private_ip')
        self.services = self.parameters.get('services')
        self.upload_file = self.parameters.get('upload')
        self.options = self.parameters.get('options') if self.parameters.get('options') else 'default'
        self.connect = ','.join(self.parameters.get('connect')) \
            if self.parameters.get('connect') and type(self.parameters.get('connect')) is list \
            else self.parameters.get('connect')
        self.private_ip_list = ','.join(self.parameters.get('private_ip_list'))
        self.use_private_ip = self.parameters.get('use_private_ip') if self.parameters.get('use_private_ip') else False

    def upload(self):
        Container(self.parameters).copy_file_to_container(self.container_name, self.upload_file, "/var/tmp")

    def copy_file(self, fl, target):
        Container(self.parameters).copy_io_to_container(self.container_name, fl, target)

    def run(self):
        working_dir = get_state_dir(self.project, self.service)
        file_output = logging.getLogger('couchformation.provisioner.output')
        file_output.propagate = False
        log_file = os.path.join(working_dir, 'provision.log')
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(CustomLogFormatter())
        file_output.addHandler(file_handler)
        file_output.setLevel(logging.DEBUG)

        command = self.command

        _command = self.resolve_variables(command)

        logger.info(f"Connecting to {self.container_name}")
        logger.debug(f"Running command: {_command}")

        file_output.info(f"{self.container_name}: [{_command}] begins")

        exit_code, output = Container(self.parameters).run_in_container(self.container_name, _command, root=self.root)

        for line in output.readlines():
            if not line:
                break
            line_out = line.decode("utf-8").strip()
            log_out = f"{self.container_name}: {line_out}"
            logger.info(log_out)
            file_output.info(log_out)

        file_output.info(f"{self.container_name}: [{_command}] complete")

        logger.info(f"Command complete for {self.container_name}")
        logger.debug(f"[{_command}] returned {exit_code} on {self.container_name}")

        file_output.removeHandler(file_handler)
        file_handler.close()
        return exit_code

    def resolve_variables(self, line: str):
        env = jinja2.Environment(undefined=jinja2.DebugUndefined)
        raw_template = env.from_string(line)
        formatted_value = raw_template.render(
            SERVICE_NAME=self.service,
            PRIVATE_IP_LIST=self.private_ip_list,
            SERVICES=self.services,
            OPTIONS=self.options,
            CONNECT_SERVICE=self.connect,
            CONNECT_IP=self.connect,
            CONNECT_LIST=self.connect if self.connect else '127.0.0.1'
        )
        return formatted_value



##
##

import logging
import botocore.exceptions
from typing import Union, List
from couchformation.aws.driver.base import CloudBase, AWSDriverError, EmptyResultSet
from couchformation.aws.driver.constants import AWSTagStruct, AWSTag

logger = logging.getLogger('couchformation.aws.driver.route')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class RouteTable(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self) -> List[dict]:
        table_list = []
        tables = []
        extra_args = {}

        try:
            while True:
                result = self.ec2_client.describe_route_tables(**extra_args)
                tables.extend(result['InternetGateways'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting Internet Gateway list: {err}")

        for table_entry in tables:
            table_block = {'owner': table_entry['OwnerId'],
                           'associations': [a for a in table_entry['Associations']],
                           'routes': [r for r in table_entry['Routes']],
                           'vpc': table_entry['VpcId'],
                           'id': table_entry['RouteTableId']}
            table_list.append(table_block)

        if len(table_list) == 0:
            raise EmptyResultSet(f"no Route Table found")
        else:
            return table_list

    def create(self, name: str, vpc_id: str, tags: dict = None) -> str:
        table_tag = AWSTagStruct.build("route-table")
        table_tag.add(AWSTag("Name", name))
        if tags:
            for k, v in tags.items():
                table_tag.add(AWSTag(k, str(v)))
        try:
            result = self.ec2_client.create_route_table(VpcId=vpc_id, TagSpecifications=[table_tag.as_dict])
        except Exception as err:
            raise AWSDriverError(f"error creating Route Table: {err}")

        return result['RouteTable']['RouteTableId']

    def delete(self, rt_id: str) -> None:
        rt = self.details(rt_id)
        if not rt:
            return
        try:
            self.ec2_client.delete_route_table(RouteTableId=rt_id)
        except Exception as err:
            raise AWSDriverError(f"error deleting Route Table: {err}")

    def get(self, name: str):
        get_filter = {
            'Name': "tag:Name",
            'Values': [
                name,
            ]
        }
        try:
            result = self.ec2_client.describe_route_tables(Filters=[get_filter])
            return result.get('RouteTables', [])[0]['RouteTableId']
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting Route Table details: {err}")

    def details(self, rt_id: str) -> Union[dict, None]:
        try:
            result = self.ec2_client.describe_route_tables(RouteTableIds=[rt_id])
            table_entry = result['RouteTables'][0]
            table_block = {'owner': table_entry['OwnerId'],
                           'associations': [a for a in table_entry['Associations']],
                           'routes': [r for r in table_entry['Routes']],
                           'vpc': table_entry['VpcId'],
                           'id': table_entry['RouteTableId']}
            return table_block
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting Route Table details: {err}")

    def associate(self, rt_id: str, subnet_id: str):
        try:
            response = self.ec2_client.associate_route_table(RouteTableId=rt_id, SubnetId=subnet_id)
            return response['AssociationId']
        except Exception as err:
            raise AWSDriverError(f"error creating Route Table: {err}")

    def add_route(self, cidr: str, ig_id: str, rt_id: str):
        try:
            response = self.ec2_client.create_route(DestinationCidrBlock=cidr, GatewayId=ig_id, RouteTableId=rt_id)
            return response['Return']
        except Exception as err:
            raise AWSDriverError(f"error creating route entry: {err}")

    def add_peer_route(self, cidr: str, peer_id: str, rt_id: str):
        try:
            response = self.ec2_client.create_route(DestinationCidrBlock=cidr, VpcPeeringConnectionId=peer_id, RouteTableId=rt_id)
            return response['Return']
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('RouteAlreadyExists'):
                return
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error creating route entry: {err}")

    def delete_route(self, cidr: str, rt_id: str):
        try:
            self.ec2_client.delete_route(DestinationCidrBlock=cidr, RouteTableId=rt_id)
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error deleting route entry: {err}")



##
##

import paramiko
import paramiko.util
import logging
import time
import socket
from typing import IO

logger = logging.getLogger('couchformation.provisioner.sftp')
logger.addHandler(logging.NullHandler())


class SSHError(Exception):
    pass


class KeepAliveFilter(logging.Filter):
    def filter(self, record):
        return record.msg.find('keepalive@lag.net') < 0


class SFTPFile(object):

    def __init__(self, ssh_key: str, ssh_user: str, hostname: str, source: str, destination: str):
        self.ssh_key = ssh_key
        self.ssh_user = ssh_user
        self.hostname = hostname
        self.source = source
        self.destination = destination

        paramiko.util.get_logger('paramiko.transport').addFilter(KeepAliveFilter())

    def ssh_connect(self, retry_count=35, factor=0.5):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.WarningPolicy())

        for retry_number in range(retry_count):
            try:
                ssh.connect(self.hostname, username=self.ssh_user, key_filename=self.ssh_key, timeout=10, auth_timeout=10, banner_timeout=10, allow_agent=False)
                ssh.get_transport().set_keepalive(5)
                return ssh
            except paramiko.ssh_exception.BadHostKeyException as err:
                raise RuntimeError(f"host key mismatch for {self.hostname}: {err}")
            except paramiko.ssh_exception.AuthenticationException as err:
                raise RuntimeError(f"failed to authenticate to {self.hostname}: {err}")
            except (paramiko.ssh_exception.SSHException, TimeoutError, socket.timeout) as err:
                n_retry = retry_number + 1
                if n_retry == retry_count:
                    raise RuntimeError(f"can not connect to {self.hostname}: {err}")
                logger.info(f"Waiting for an SSH connection to {self.hostname}")
                wait = factor
                wait *= n_retry
                time.sleep(wait)

    def upload(self, connect_retry=35, op_retry=10, factor=0.5):
        ssh = self.ssh_connect(retry_count=connect_retry, factor=factor)
        sftp_client = ssh.open_sftp()

        for retry_number in range(op_retry):
            try:
                sftp_client.put(self.source, self.destination)
            except Exception as err:
                n_retry = retry_number + 1
                if n_retry == op_retry:
                    raise RuntimeError(f"upload failed on {self.hostname}: {err}")
                logger.info(f"Retrying upload on {self.hostname}")
                wait = factor
                wait *= n_retry
                time.sleep(wait)

        sftp_client.close()

    def download(self, connect_retry=35, op_retry=10, factor=0.5):
        ssh = self.ssh_connect(retry_count=connect_retry, factor=factor)
        sftp_client = ssh.open_sftp()

        for retry_number in range(op_retry):
            try:
                sftp_client.get(self.source, self.destination)
            except Exception as err:
                n_retry = retry_number + 1
                if n_retry == op_retry:
                    raise RuntimeError(f"upload failed on {self.hostname}: {err}")
                logger.info(f"Retrying upload on {self.hostname}")
                wait = factor
                wait *= n_retry
                time.sleep(wait)

        sftp_client.close()


class SFTPFileObject(object):

    def __init__(self, ssh_key: str, ssh_user: str, hostname: str, fl: IO, destination: str):
        self.ssh_key = ssh_key
        self.ssh_user = ssh_user
        self.hostname = hostname
        self.fl = fl
        self.destination = destination

        paramiko.util.get_logger('paramiko.transport').addFilter(KeepAliveFilter())

    def ssh_connect(self, retry_count=35, factor=0.5):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.WarningPolicy())

        for retry_number in range(retry_count):
            try:
                ssh.connect(self.hostname, username=self.ssh_user, key_filename=self.ssh_key, timeout=10, auth_timeout=10, banner_timeout=10, allow_agent=False)
                ssh.get_transport().set_keepalive(5)
                return ssh
            except paramiko.ssh_exception.BadHostKeyException as err:
                raise RuntimeError(f"host key mismatch for {self.hostname}: {err}")
            except paramiko.ssh_exception.AuthenticationException as err:
                raise RuntimeError(f"failed to authenticate to {self.hostname}: {err}")
            except (paramiko.ssh_exception.SSHException, TimeoutError, socket.timeout) as err:
                n_retry = retry_number + 1
                if n_retry == retry_count:
                    raise RuntimeError(f"can not connect to {self.hostname}: {err}")
                logger.info(f"Waiting for an SSH connection to {self.hostname}")
                wait = factor
                wait *= n_retry
                time.sleep(wait)

    def copy_file(self, connect_retry=35, op_retry=10, factor=0.5):
        ssh = self.ssh_connect(retry_count=connect_retry, factor=factor)
        sftp_client = ssh.open_sftp()

        for retry_number in range(op_retry):
            try:
                self.fl.seek(0)
                f = sftp_client.open(self.destination, 'wb')
                f.write(self.fl.read())
                f.close()
            except Exception as err:
                n_retry = retry_number + 1
                if n_retry == op_retry:
                    raise RuntimeError(f"upload failed on {self.hostname}: {err}")
                logger.info(f"Retrying upload on {self.hostname}")
                wait = factor
                wait *= n_retry
                time.sleep(wait)

        sftp_client.close()



##
##

import multiprocessing.queues


class TaskQueue(multiprocessing.queues.SimpleQueue):

    def __init__(self, *args, **kwargs):
        kwargs['ctx'] = multiprocessing.get_context('fork')
        super().__init__(*args, **kwargs)

    def send_task(self, module, instance, method, *args, **kwargs):
        m = __import__(module, fromlist=[""])
        i = getattr(m, instance)
        obj = i(*args, **kwargs)
        f = getattr(obj, method)
        self.put(f)

    def get_task(self):
        return self.get()



##
##

import logging

logger = logging.getLogger('couchformation.null')
logger.addHandler(logging.NullHandler())


class NullClass(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters

    def null_f(self):
        logger.debug(f"Null call for name {self.parameters.get('name')}")
        pass



##
##

import os
import logging
import random
import string
from itertools import cycle
from typing import List
from couchformation.network import NetworkDriver
from couchformation.aws.driver.network import Network, Subnet
from couchformation.aws.driver.sshkey import SSHKey
from couchformation.aws.driver.gateway import InternetGateway
from couchformation.aws.driver.nsg import SecurityGroup
from couchformation.aws.driver.route import RouteTable
from couchformation.aws.driver.dns import DNS
import couchformation.aws.driver.constants as C
from couchformation.config import get_state_file, get_state_dir, PortSettingSet, PortSettings, State
from couchformation.deployment import MetadataManager
from couchformation.ssh import SSHUtil
from couchformation.exception import FatalError
from couchformation.kvdb import KeyValueStore
from couchformation.util import FileManager, synchronize, UUIDGen, parameter_to_dict, csv_dict_concat, dump_class_variables

logger = logging.getLogger('couchformation.aws.network')
logger.addHandler(logging.NullHandler())


class AWSNetworkError(FatalError):
    pass


class AWSNetwork(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.project = parameters.get('project')
        self.region = parameters.get('region')
        self.auth_mode = parameters.get('auth_mode')
        self.profile = parameters.get('profile')
        self.ssh_key = parameters.get('ssh_key')
        self.cloud = parameters.get('cloud')
        self.domain = parameters.get('domain')
        self.hosted_zone = parameters.get('hosted_zone')
        self.provider_id = parameters.get('provider_id')
        self.tags = parameters.get('tags') if parameters.get('tags') else ''
        self.allow = parameters.get('allow') if parameters.get('allow') else "0.0.0.0/0"
        self.build_ports = PortSettingSet().create().items()

        filename = get_state_file(self.project, f"network-{self.region}")

        try:
            state_dir = get_state_dir(self.project, f"network-{self.region}")
            if not os.path.exists(state_dir):
                FileManager().make_dir(state_dir)
        except Exception as err:
            raise AWSNetworkError(f"can not create state dir: {err}")

        document = f"network:{self.cloud}"
        self.state = KeyValueStore(filename, document)

        self.aws_network = Network(self.parameters)

        self.tags = csv_dict_concat({"Project": self.project}, self.tags)
        self.tags = csv_dict_concat({"Service": self.name}, self.tags)

        project_uid = MetadataManager(self.project).project_uid
        self.asset_prefix = f"cf-{project_uid}"
        self.vpc_name = f"{self.asset_prefix}-vpc"
        self.ig_name = f"{self.asset_prefix}-gw"
        self.rt_name = f"{self.asset_prefix}-rt"
        self.sg_name = f"{self.asset_prefix}-sg"
        self.key_name = f"{self.asset_prefix}-key"

    def check_state(self):
        for n, zone_state in reversed(list(enumerate(self.state.list_get('zone')))):
            subnet_id = zone_state[2]
            result = Subnet(self.parameters).details(subnet_id)
            if result is None:
                logger.warning(f"Removing stale state entry for subnet {subnet_id}")
                self.state.list_remove('zone', zone_state[0])

        if self.state.get('route_table_id'):
            result = RouteTable(self.parameters).details(self.state['route_table_id'])
            if result is None:
                logger.warning(f"Removing stale state entry for route table {self.state['route_table_id']}")
                del self.state['route_table_id']
        else:
            result = RouteTable(self.parameters).get(self.rt_name)
            if result:
                logger.warning(f"Importing orphaned entry for route table {result}")
                self.state['route_table_id'] = result

        if self.state.get('internet_gateway_id'):
            result = InternetGateway(self.parameters).details(self.state['internet_gateway_id'])
            if result is None:
                logger.warning(f"Removing stale state entry for gateway {self.state['internet_gateway_id']}")
                del self.state['internet_gateway_id']
        else:
            result = InternetGateway(self.parameters).get(self.ig_name)
            if result:
                logger.warning(f"Importing orphaned entry for internet gateway {result}")
                self.state['internet_gateway_id'] = result

        if self.state.get('security_group_id'):
            result = SecurityGroup(self.parameters).details(self.state['security_group_id'])
            if result is None:
                logger.warning(f"Removing stale state entry for security group {self.state['security_group_id']}")
                del self.state['security_group_id']
        else:
            result = SecurityGroup(self.parameters).get(self.sg_name)
            if result:
                logger.warning(f"Importing orphaned entry for security group {result}")
                self.state['security_group_id'] = result

        for build_port_cfg in self.build_ports:
            build_name = build_port_cfg.build
            state_key_name = f"{build_name}_security_group_id"
            build_sg_name = f"{self.asset_prefix}-{build_name}-sg"
            if self.state.get(state_key_name):
                result = SecurityGroup(self.parameters).details(self.state[state_key_name])
                if result is None:
                    logger.warning(f"Removing stale state entry for security group {self.state[state_key_name]}")
                    del self.state[state_key_name]
            else:
                result = SecurityGroup(self.parameters).get(build_sg_name)
                if result:
                    logger.warning(f"Importing orphaned entry for security group {result}")
                    self.state[state_key_name] = result

        if self.state.get('win_security_group_id'):
            result = SecurityGroup(self.parameters).details(self.state['win_security_group_id'])
            if result is None:
                logger.warning(f"Removing stale state entry for security group {self.state['win_security_group_id']}")
                del self.state['win_security_group_id']
        else:
            win_sg_name = f"{self.asset_prefix}-win-sg"
            result = SecurityGroup(self.parameters).get(win_sg_name)
            if result:
                logger.warning(f"Importing orphaned entry for security group {result}")
                self.state['win_security_group_id'] = result

        for group_sg_key in self.state.key_match('.*_group_.*_sg_id'):
            if self.state.get(group_sg_key):
                result = SecurityGroup(self.parameters).details(self.state[group_sg_key])
                if result is None:
                    logger.warning(f"Removing stale state entry for security group {self.state[group_sg_key]}")
                    del self.state[group_sg_key]

        for n, sg_group in enumerate(SecurityGroup(self.parameters).search(f"{self.asset_prefix}-*-sg")):
            sg_group_id = sg_group.get('id')
            service = sg_group.get('Service', 'import')
            group = sg_group.get('Group', n)
            state_key_name = f"{service}_group_{group}_sg_id"
            if not self.state.value_match(sg_group_id):
                logger.warning(f"Importing orphaned entry for security group {sg_group_id}")
                self.state[state_key_name] = sg_group_id

        if self.state.get('vpc_id'):
            result = Network(self.parameters).details(self.state['vpc_id'])
            if result is None:
                logger.warning(f"Removing stale state entry for network {self.state['vpc_id']}")
                del self.state['vpc_id']
                del self.state['vpc_cidr']
                del self.state['zone']
        if self.state.get('ssh_key'):
            result = SSHKey(self.parameters).details(self.state['ssh_key'])
            if result is None:
                logger.warning(f"Removing stale state entry for SSH key {self.state['ssh_key']}")
                del self.state['ssh_key']
        if self.state.get('public_hosted_zone'):
            result = DNS(self.parameters).details(self.state['public_hosted_zone'])
            if result is None:
                logger.warning(f"Removing stale state entry for public hosted domain {self.state['public_hosted_zone']}")
                del self.state['public_hosted_zone']
        if self.state.get('private_hosted_zone'):
            result = DNS(self.parameters).details(self.state['private_hosted_zone'])
            if result is None:
                logger.warning(f"Removing stale state entry for private hosted domain {self.state['private_hosted_zone']}")
                del self.state['private_hosted_zone']

    @synchronize()
    def create_vpc(self):
        self.check_state()
        cidr_util = NetworkDriver()

        for net in self.aws_network.cidr_list:
            cidr_util.add_network(net)

        zone_list = self.aws_network.zones()

        ssh_pub_key_text = SSHUtil().get_ssh_public_key(self.ssh_key)

        account_id = self.aws_network.account_id
        logger.info(f"Creating VPC as Account ID {account_id}")
        logger.info(f"VPC Region: {self.region}")

        try:
            self.state['state'] = State.DEPLOYING.value
            logger.debug(f"AWS VPC create input variables:\n{dump_class_variables(vars(self))}")

            if not self.state.get('account_id'):
                self.state['account_id'] = account_id

            if not self.state.get('vpc_id'):
                vpc_cidr = cidr_util.get_next_network()
                vpc_id = Network(self.parameters).create(self.vpc_name, vpc_cidr, parameter_to_dict(self.tags))
                Network(self.parameters).enable_dns_hostnames(vpc_id)
                self.state['vpc_id'] = vpc_id
                self.state['vpc_cidr'] = vpc_cidr
                logger.info(f"Created VPC {vpc_id}")
                logger.info(f"VPC CIDR: {vpc_cidr}")
            else:
                vpc_id = self.state.get('vpc_id')
                vpc_cidr = self.state.get('vpc_cidr')
                cidr_util.set_active_network(vpc_cidr)

            subnet_list = list(cidr_util.get_next_subnet())
            subnet_cycle = cycle(subnet_list[1:])

            if not self.state.get('security_group_id'):
                sg_id = SecurityGroup(self.parameters).create(self.sg_name,
                                                              f"Couch Formation project {self.project}",
                                                              vpc_id,
                                                              parameter_to_dict(self.tags))
                SecurityGroup(self.parameters).add_ingress(sg_id, "-1", 0, 0, vpc_cidr)
                SecurityGroup(self.parameters).add_ingress(sg_id, "tcp", 22, 22, self.allow)
                self.state['security_group_id'] = sg_id
                logger.info(f"Created security group {sg_id}")

            if not self.state.get('ssh_key'):
                ssh_key_name = SSHKey(self.parameters).create(self.key_name, ssh_pub_key_text, parameter_to_dict(self.tags))
                self.state['ssh_key'] = ssh_key_name
                logger.info(f"Created SSH Key {ssh_key_name}")

            if not self.state.get('internet_gateway_id'):
                ig_id = InternetGateway(self.parameters).create(self.ig_name, vpc_id, parameter_to_dict(self.tags))
                self.state['internet_gateway_id'] = ig_id
                logger.info(f"Created internet gateway {ig_id}")
            else:
                ig_id = self.state.get('internet_gateway_id')

            if not self.state.get('route_table_id'):
                rt_id = RouteTable(self.parameters).create(self.rt_name, vpc_id, parameter_to_dict(self.tags))
                self.state['route_table_id'] = rt_id
                RouteTable(self.parameters).add_route("0.0.0.0/0", ig_id, rt_id)
                logger.info(f"Created route table {rt_id}")
            else:
                rt_id = self.state.get('route_table_id')

            for n, zone in enumerate(zone_list):
                if self.state.list_exists('zone', zone):
                    continue
                while True:
                    network_cidr = next(subnet_cycle)
                    if not self.state.list_exists('zone', network_cidr):
                        break
                subnet_name = f"{self.asset_prefix}-subnet-{n+1:02d}"
                subnet_id = Subnet(self.parameters).create(subnet_name, vpc_id, zone, network_cidr, parameter_to_dict(self.tags))
                RouteTable(self.parameters).associate(rt_id, subnet_id)
                self.state.list_add('zone', zone, network_cidr, subnet_id)
                logger.info(f"Created subnet {subnet_id} in zone {zone}")

            logger.debug(f"DNS domain name: {self.domain}")
            if self.domain and not self.state.get('domain'):
                domain_prefix = ''.join(random.choice(string.ascii_lowercase) for _ in range(7))
                domain_name = f"{domain_prefix}.{self.domain}"
                self.state['domain'] = domain_name
                logger.info(f"Generated project domain {domain_name}")
            elif self.state.get('domain'):
                domain_name = self.state.get('domain')
                logger.info(f"Using existing domain {domain_name}")
            else:
                domain_name = None

            if domain_name and not self.state.get('public_hosted_zone'):
                domain_id = DNS(self.parameters).create(domain_name)
                self.state['public_hosted_zone'] = domain_id
                logger.info(f"Created public hosted zone {domain_id} for domain {domain_name}")

            if domain_name and not self.state.get('private_hosted_zone'):
                domain_id = DNS(self.parameters).create(domain_name, vpc_id, self.region)
                self.state['private_hosted_zone'] = domain_id
                logger.info(f"Created private hosted zone {domain_id} for domain {domain_name}")

            if self.state.get('public_hosted_zone') and not self.state['parent_hosted_zone']:
                parent_domain = '.'.join(domain_name.split('.')[1:])
                parent_id = DNS(self.parameters).zone_id(parent_domain)
                if parent_id:
                    ns_names = DNS(self.parameters).record_sets(self.state['public_hosted_zone'], 'NS')
                    DNS(self.parameters).add_record(parent_id, domain_name, ns_names, 'NS')
                    self.state['parent_hosted_zone'] = parent_id
                    self.state['parent_zone_ns_records'] = ','.join(ns_names)
                    logger.info(f"Added {len(ns_names)} NS record(s) to domain {parent_domain}")

            self.state['state'] = State.DEPLOYED.value
        except Exception as err:
            raise AWSNetworkError(f"Error creating VPC: {err}")

    @synchronize()
    def create_build_sg(self, build_name: str):
        vpc_id = self.vpc_id
        for build_port_cfg in self.build_ports:
            if build_port_cfg.build != build_name:
                continue
            state_key_name = f"{build_name}_security_group_id"
            build_sg_name = f"{self.asset_prefix}-{build_name}-sg"
            if not self.state.get(state_key_name):
                build_sg_id = SecurityGroup(self.parameters).create(build_sg_name, f"Couch Formation build type {build_name}", vpc_id)
                for begin, end in build_port_cfg.tcp_as_tuple():
                    SecurityGroup(self.parameters).add_ingress(build_sg_id, "tcp", begin, end, self.allow)
                for begin, end in build_port_cfg.udp_as_tuple():
                    SecurityGroup(self.parameters).add_ingress(build_sg_id, "udp", begin, end, self.allow)
                self.state[state_key_name] = build_sg_id
                logger.info(f"Created {build_name} build security group {build_sg_id}")
            else:
                build_sg_id = self.state.get(state_key_name)
            return build_sg_id

    @synchronize()
    def create_win_sg(self):
        vpc_id = self.vpc_id
        if not self.state.get('win_security_group_id'):
            win_sg_name = f"{self.asset_prefix}-win-sg"
            win_sg_id = SecurityGroup(self.parameters).create(win_sg_name, "Couch Formation Windows OS ports", vpc_id)
            SecurityGroup(self.parameters).add_ingress(win_sg_id, "tcp", 3389, 3389, self.allow)
            SecurityGroup(self.parameters).add_ingress(win_sg_id, "tcp", 5985, 5986, self.allow)
            self.state['win_security_group_id'] = win_sg_id
            logger.info(f"Created win security group {win_sg_id}")
        else:
            win_sg_id = self.state.get('win_security_group_id')
        return win_sg_id

    @synchronize()
    def create_node_group_sg(self, service: str, group: int, ports: List[str]):
        vpc_id = self.vpc_id
        state_key_name = f"{service}_group_{group}_sg_id"
        service_code = UUIDGen().text_hash(f"{service}-group-{group}")
        build_sg_name = f"{self.asset_prefix}-{service_code}-sg"
        if not self.state.get(state_key_name):
            port_cfg = PortSettings().create(self.name, ports)
            tags = {'Service': service, 'Group': group}
            port_sg_id = SecurityGroup(self.parameters).create(build_sg_name, f"Couch Formation service {service} group {group}", vpc_id, tags=tags)
            for begin, end in port_cfg.tcp_as_tuple():
                SecurityGroup(self.parameters).add_ingress(port_sg_id, "tcp", begin, end, self.allow)
            for begin, end in port_cfg.udp_as_tuple():
                SecurityGroup(self.parameters).add_ingress(port_sg_id, "udp", begin, end, self.allow)
            self.state[state_key_name] = port_sg_id
            logger.info(f"Created service group security group {port_sg_id}")
        else:
            port_sg_id = self.state.get(state_key_name)
        return port_sg_id

    @synchronize()
    def peer_vpc(self):
        self.check_state()
        vpc_id = self.state.get('vpc_id')
        rt_id = self.state.get('route_table_id')

        if self.provider_id:
            pcx_id = self.provider_id
            self.state['peering_id'] = pcx_id
            peering = Network(self.parameters).peering_get(pcx_id)
            if peering.get('status') == 'pending-acceptance':
                cidr = peering.get('cidr')
                self.state['peer_cidr'] = cidr
                Network(self.parameters).peering_accept(pcx_id)
                logger.info(f"Accepted peering request {pcx_id} with CIDR {cidr}")
                RouteTable(self.parameters).add_peer_route(cidr, pcx_id, rt_id)
                logger.info(f"Added route for {cidr} to route table {rt_id}")

        if self.hosted_zone:
            self.state['peer_hosted_zone'] = self.hosted_zone
            DNS(self.parameters).associate(self.hosted_zone, vpc_id, self.region)
            logger.info(f"Associated hosted zone {self.hosted_zone} with VPC {vpc_id}")

    @synchronize()
    def unpeer_vpc(self):
        self.check_state()
        vpc_id = self.state.get('vpc_id')
        rt_id = self.state.get('route_table_id')

        if self.state.get('peering_id'):
            peering = Network(self.parameters).peering_get(self.state['peering_id'])
            if peering.get('status') != 'pending-acceptance' and peering.get('status') != 'failed' and peering.get('status') != 'rejected' and peering.get('status') != 'deleted':
                Network(self.parameters).peering_delete(self.state['peering_id'])
                logger.info(f"Deleted peering ID {self.state['peering_id']}")
                del self.state['peering_id']

        if self.state.get('peer_cidr'):
            RouteTable(self.parameters).delete_route(self.state['peer_cidr'], rt_id)
            logger.info(f"Deleted route for {self.state['peer_cidr']} in route table {rt_id}")
            del self.state['peer_cidr']

        if self.state.get('peer_hosted_zone'):
            DNS(self.parameters).disassociate(self.state['peer_hosted_zone'], vpc_id, self.region)
            logger.info(f"Disassociated hosted zone {self.state['peer_hosted_zone']} from VPC {vpc_id}")
            del self.state['peer_hosted_zone']

    @synchronize()
    def destroy_vpc(self):
        if self.state.list_len('services') > 0:
            logger.info(f"Active services, leaving project network in place")
            return

        self.check_state()
        try:
            self.state['state'] = State.DESTROYING.value

            for n, zone_state in reversed(list(enumerate(self.state.list_get('zone')))):
                subnet_id = zone_state[2]
                Subnet(self.parameters).delete(subnet_id)
                self.state.list_remove('zone', zone_state[0])
                logger.info(f"Removed subnet {subnet_id}")

            if self.state.get('route_table_id'):
                rt_id = self.state.get('route_table_id')
                RouteTable(self.parameters).delete(rt_id)
                del self.state['route_table_id']
                logger.info(f"Removed route table {rt_id}")

            if self.state.get('internet_gateway_id'):
                ig_id = self.state.get('internet_gateway_id')
                InternetGateway(self.parameters).delete(ig_id)
                del self.state['internet_gateway_id']
                logger.info(f"Removing internet gateway {ig_id}")

            if self.state.get('security_group_id'):
                sg_id = self.state.get('security_group_id')
                SecurityGroup(self.parameters).delete(sg_id)
                del self.state['security_group_id']
                logger.info(f"Removing security group {sg_id}")

            for build_port_cfg in self.build_ports:
                build_name = build_port_cfg.build
                state_key_name = f"{build_name}_security_group_id"
                if self.state.get(state_key_name):
                    sg_id = self.state.get(state_key_name)
                    SecurityGroup(self.parameters).delete(sg_id)
                    del self.state[state_key_name]
                    logger.info(f"Removing security group {sg_id}")

            if self.state.get('win_security_group_id'):
                sg_id = self.state.get('win_security_group_id')
                SecurityGroup(self.parameters).delete(sg_id)
                del self.state['win_security_group_id']
                logger.info(f"Removing security group {sg_id}")

            for group_sg_key in self.state.key_match('.*_group_.*_sg_id'):
                if self.state.get(group_sg_key):
                    sg_id = self.state.get(group_sg_key)
                    SecurityGroup(self.parameters).delete(sg_id)
                    del self.state[group_sg_key]
                    logger.info(f"Removing security group {sg_id}")

            if self.state.get('vpc_id'):
                vpc_id = self.state.get('vpc_id')
                Network(self.parameters).delete(vpc_id)
                del self.state['vpc_id']
                del self.state['vpc_cidr']
                logger.info(f"Removing VPC {vpc_id}")

            if self.state.get('ssh_key'):
                ssh_key_name = self.state.get('ssh_key')
                instances = SSHKey(self.parameters).instances_by_key(ssh_key_name)
                if len(instances) > 0:
                    logger.warning(f"SSH key {ssh_key_name} in use, key will not be deleted from AWS")
                else:
                    logger.info(f"Deleting SSH key pair {ssh_key_name}")
                    SSHKey(self.parameters).delete(ssh_key_name)
                del self.state['ssh_key']
                logger.info(f"Removing key pair {ssh_key_name} from project")

            if self.state.get('parent_hosted_zone') and self.state.get('domain'):
                ns_names = self.state['parent_zone_ns_records'].split(',')
                DNS(self.parameters).delete_record(self.state['parent_hosted_zone'], self.state['domain'], ns_names, 'NS')
                del self.state['parent_hosted_zone']
                del self.state['parent_zone_ns_records']
                logger.info(f"Removing NS records for domain {self.state['domain']}")

            if self.state.get('public_hosted_zone'):
                domain_id = self.state.get('public_hosted_zone')
                DNS(self.parameters).delete(domain_id)
                del self.state['public_hosted_zone']
                logger.info(f"Removing public hosted zone {domain_id}")

            if self.state.get('private_hosted_zone'):
                domain_id = self.state.get('private_hosted_zone')
                DNS(self.parameters).delete(domain_id)
                del self.state['private_hosted_zone']
                logger.info(f"Removing private hosted zone {domain_id}")

            if self.state.get('domain'):
                domain_name = self.state.get('domain')
                del self.state['domain']
                logger.info(f"Removing project domain {domain_name}")

            self.state['state'] = State.IDLE.value
        except Exception as err:
            raise AWSNetworkError(f"Error removing VPC: {err}")

    def create(self):
        logger.info(f"Creating cloud network for {self.project} in {C.CLOUD_KEY.upper()}")
        self.create_vpc()

    def destroy(self):
        logger.info(f"Removing cloud network for {self.project} in {C.CLOUD_KEY.upper()}")
        self.destroy_vpc()

    def get(self, key):
        return self.state.get(key)

    @property
    def ssh_key_id(self):
        return self.state.get('ssh_key')

    @property
    def security_group_id(self):
        return self.state.get('security_group_id')

    def build_security_group_id(self, build_name: str):
        state_key_name = f"{build_name}_security_group_id"
        return self.state.get(state_key_name)

    @property
    def win_security_group_id(self):
        return self.state.get('win_security_group_id')

    @property
    def zones(self):
        return self.state.list_get('zone')

    @property
    def domain_name(self):
        return self.state.get('domain')

    @property
    def public_zone(self):
        return self.state.get('public_hosted_zone')

    @property
    def private_zone(self):
        return self.state.get('private_hosted_zone')

    @property
    def vpc_id(self):
        return self.state.get('vpc_id')

    def add_service(self, name):
        self.state.list_add('services', name)

    def remove_service(self, name):
        self.state.list_remove('services', name)



##
##

import logging
from azure.core.exceptions import ResourceNotFoundError
from couchformation.azure.driver.base import CloudBase, AzureDriverError

logger = logging.getLogger('couchformation.azure.driver.private_dns')
logger.addHandler(logging.NullHandler())
logging.getLogger("azure").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class PrivateDNS(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create(self, domain: str, resource_group: str) -> str:
        parameters = {
            'location': 'global'
        }

        try:
            request = self.private_dns_client.private_zones.begin_create_or_update(resource_group, domain, parameters)
            request.wait()
            result = request.result()
            return result.name
        except Exception as err:
            raise AzureDriverError(f"error creating zone: {err}")

    def details(self, domain: str):
        zones = self.list_zones()
        return next((zone for zone in zones if zone['name'].startswith(domain)), None)

    def zone_name(self, domain: str):
        zone = self.details(domain)
        return zone['name'] if zone else None

    def zone_rg(self, domain: str):
        zone = self.details(domain)
        return zone['resource_group'] if zone else None

    def list_zones(self):
        zone_list = []

        try:
            zones = self.private_dns_client.private_zones.list()
        except Exception as err:
            raise AzureDriverError(f"error getting zones: {err}")

        for group in list(zones):
            zone_data = {
                'id': group.id,
                'etag': group.etag,
                'resource_group': group.id.split('/')[4],
                'name': group.name,
                'location': group.location,
            }
            zone_list.append(zone_data)

        return zone_list

    def record_sets(self, name: str, r_type: str, resource_group: str):
        record_data = []

        try:
            records = self.private_dns_client.record_sets.list_by_type(resource_group, name, r_type)
        except Exception as err:
            raise AzureDriverError(f"error getting resource records: {err}")

        def _list_iter(items):
            if items:
                return list(items)
            else:
                return []

        for group in list(records):
            record = {
                'id': group.id,
                'name': group.name,
                'a_records': [r.ipv4_address for r in _list_iter(group.a_records)],
                'aaaa_records': [r.ipv6_address for r in _list_iter(group.aaaa_records)],
                'mx_records': [r.exchange for r in _list_iter(group.mx_records)],
                'ptr_records': [r.ptrdname for r in _list_iter(group.ptr_records)],
                'srv_records': [f"{r.priority} {r.weight} {r.port} {r.target}" for r in _list_iter(group.srv_records)],
                'txt_records': [r.value for r in _list_iter(group.txt_records)],
                'cname_records': [r.cname for r in _list_iter(group.cname_record)],
                'soa_records': [f"{r.host} {r.email} {r.serial_number} {r.refresh_time} {r.retry_time} {r.expire_time} {r.minimum_ttl}" for r in _list_iter(group.soa_record)]
            }
            record_data.append(record)

        return record_data[0].get(f"{r_type.lower()}_records", [])

    def delete(self, name: str, resource_group: str):
        try:
            request = self.private_dns_client.private_zones.begin_delete(resource_group, name)
            request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error deleting zone: {err}")

    @staticmethod
    def parameter_gen(values: list, record_type: str = 'A', ttl: int = 300):
        if record_type == 'A':
            return {"properties": {"aRecords": [{"ipv4Address": v} for v in values], "TTL": ttl}}
        elif record_type == 'AAAA':
            return {"properties": {"aaaaRecords": [{"ipv6Address": v} for v in values], "TTL": ttl}}
        elif record_type == 'CNAME':
            return {"properties": {"cnameRecord": {"cname": values[0]}, "TTL": ttl}}
        elif record_type == 'MX':
            return {"properties": {"mxRecords": [{"exchange": v, "preference": n} for n, v in enumerate(values)], "TTL": ttl}}
        elif record_type == 'PTR':
            return {"properties": {"ptrRecords": [{"ptrdname": v} for v in values], "TTL": ttl}}
        elif record_type == 'SRV':
            return {"properties": {"srvRecords": [{"priority": v.split()[0], "weight": v.split()[1], "port": v.split()[2], "target": v.split()[3]} for v in values], "TTL": ttl}}
        elif record_type == 'TXT':
            return {"properties": {"txtRecords": [{"value": values}], "TTL": ttl}}

    def add_record(self, zone_name: str, name: str, values: list, resource_group: str, record_type: str = 'A', ttl: int = 300):
        relative_name = name.split('.')[:1][0]
        try:
            parameters = self.parameter_gen(values, record_type, ttl)
            kwargs = dict(
                resource_group_name=resource_group,
                private_zone_name=zone_name,
                relative_record_set_name=relative_name,
                record_type=record_type,
                parameters=parameters
            )
            response = self.private_dns_client.record_sets.create_or_update(**kwargs)
            return response.id
        except Exception as err:
            raise AzureDriverError(f"error creating zone: {err}")

    def delete_record(self, zone_name: str, name: str, resource_group: str, record_type: str = 'A'):
        relative_name = name.split('.')[:1][0]
        try:
            self.private_dns_client.record_sets.delete(resource_group, zone_name, record_type, relative_name)
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error deleting record: {err}")

    def vpc_link(self, zone_name: str, name: str, vnet_id: str, resource_group: str):
        parameters = {
            "location": "Global",
            "properties": {
                "virtualNetwork": {
                    "id": vnet_id
                },
                "registrationEnabled": False
            }
        }
        try:
            request = self.private_dns_client.virtual_network_links.begin_create_or_update(resource_group, zone_name, name, parameters)
            request.wait()
        except Exception as err:
            raise AzureDriverError(f"error creating vnet link: {err}")

    def vpc_unlink(self, zone_name: str, name: str, resource_group: str):
        try:
            request = self.private_dns_client.virtual_network_links.begin_delete(resource_group, zone_name, name)
            request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error removing vnet link: {err}")

    def vpc_link_details(self, zone_name: str, name: str, resource_group: str):
        try:
            result = self.private_dns_client.virtual_network_links.get(resource_group, zone_name, name)
            return result
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting vnet link: {err}")



##
##

import attr
import os
from pathlib import Path

CLOUD_KEY = "azure"


def get_auth_directory():
    return os.path.join(Path.home(), '.azure')


def get_config_default():
    return os.path.join(get_auth_directory(), 'clouds.config')


def get_config_main():
    return os.path.join(get_auth_directory(), 'config')


@attr.s
class AzureDiskTypes(object):
    disk_type_list = [
        {
            "type": 'Premium_LRS'
        },
        {
            "type": 'UltraSSD_LRS'
        }
    ]


@attr.s
class AzureDiskTiers(object):
    disk_tier_list = [
        {
            "disk_size": "4",
            "disk_tier": "P2",
            "disk_iops": "1200"
        },
        {
            "disk_size": "8",
            "disk_tier": "P3",
            "disk_iops": "2400"
        },
        {
            "disk_size": "16",
            "disk_tier": "P4",
            "disk_iops": "4800"
        },
        {
            "disk_size": "32",
            "disk_tier": "P6",
            "disk_iops": "9600"
        },
        {
            "disk_size": "64",
            "disk_tier": "P10",
            "disk_iops": "16000"
        },
        {
            "disk_size": "128",
            "disk_tier": "P15",
            "disk_iops": "16000"
        },
        {
            "disk_size": "256",
            "disk_tier": "P20",
            "disk_iops": "16000"
        },
        {
            "disk_size": "512",
            "disk_tier": "P30",
            "disk_iops": "16000"
        },
        {
            "disk_size": "1024",
            "disk_tier": "P40",
            "disk_iops": "16000"
        },
        {
            "disk_size": "2048",
            "disk_tier": "P50",
            "disk_iops": "16000"
        },
        {
            "disk_size": "4096",
            "disk_tier": "P50",
            "disk_iops": "16000"
        },
        {
            "disk_size": "8192",
            "disk_tier": "P80",
            "disk_iops": "16000"
        },
        {
            "disk_size": "16384",
            "disk_tier": "P80",
            "disk_iops": "16000"
        }
    ]


@attr.s
class AzureImagePublishers(object):
    publishers = [
        {
            "name": "Canonical",
            "offer_match": r"^0001-com-ubuntu-server-.*",
            "description": "Ubuntu Linux",
            "os_id": "ubuntu",
            "user": "ubuntu",
            "sku_match": "^(.+?)-lts-gen2$"
        },
        {
            "name": "OpenLogic",
            "offer_match": r"^CentOS$",
            "description": "CentOS Linux",
            "os_id": "centos",
            "user": "centos",
            "sku_match": r"^(.+?)_[0-9]-gen2$"
        },
        {
            "name": "RedHat",
            "offer_match": r"^RHEL$",
            "description": "RedHat Linux",
            "os_id": "rhel",
            "user": "rhel",
            "sku_match": r"^(.+?)[0-9]-gen2$"
        },
        {
            "name": "SUSE",
            "offer_match": r"^sles-(.+?)-sp[0-9]$",
            "description": "Suse Linux",
            "os_id": "sles",
            "user": "sles",
            "sku_match": "^gen2$"
        },
        {
            "name": "SUSE",
            "offer_match": r"^opensuse-leap-(.+?)-[0-9]$",
            "description": "Suse Linux",
            "os_id": "opensuse-leap",
            "user": "sles",
            "sku_match": "^gen2$"
        },
        {
            "name": "Debian",
            "offer_match": r"^debian-[0-9]*$",
            "description": "Debian 10 and later",
            "os_id": "debian",
            "user": "debian",
            "sku_match": "^(.+?)-gen2$"
        },
        {
            "name": "MicrosoftWindowsServer",
            "offer_match": r"^WindowsServer$",
            "description": "Windows Server",
            "os_id": "windows",
            "user": "adminuser",
            "sku_match": "^(.+?)-datacenter-g[2e].*$"
        },
    ]


@attr.s
class ComputeTypes(object):
    size_family = set('DEFL')
    size_features = set('tlmsa')
    vmp_features = set('tlms')
    size_storage = set('s')
    size_versions = set('2345')
    vmp_versions = set('345')


@attr.s
class StorageTierMap(object):
    size_in_gb = {
        "4": ["P2", "P3", "P4", "P6", "P10", "P15", "P20", "P30", "P40", "P50"],
        "8": ["P3", "P4", "P6", "P10", "P15", "P20", "P30", "P40", "P50"],
        "16": ["P4", "P6", "P10", "P15", "P20", "P30", "P40", "P50"],
        "32": ["P6", "P10", "P15", "P20", "P30", "P40", "P50"],
        "64": ["P10", "P15", "P20", "P30", "P40", "P50"],
        "128": ["P15", "P20", "P30", "P40", "P50"],
        "256": ["P20", "P30", "P40", "P50"],
        "512": ["P30", "P40", "P50"],
        "1024": ["P40", "P50"],
        "2048": ["P50"],
        "4096": ["P50"],
        "8192": ["P70", "P80"],
        "16384": ["P80"],
        "32768": ["P80"],
    }



##
##

import re
import logging
from typing import Union, List
from couchformation.azure.driver.base import CloudBase, AzureDriverError
from couchformation.azure.driver.constants import ComputeTypes
import couchformation.constants as C

logger = logging.getLogger('couchformation.azure.driver.machine')
logger.addHandler(logging.NullHandler())
logging.getLogger("azure").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class MachineType(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self, location: str, virtualization: bool = False) -> list:
        machine_type_list = []

        try:
            machine_list = self.compute_client.virtual_machine_sizes.list(location=location)
        except Exception as err:
            raise AzureDriverError(f"error listing machine types: {err}")

        for machine in list(machine_list):
            match = re.search(r"Standard_([A-Z])([0-9]*)([a-z]*)_v([0-9])", machine.name)
            if not match:
                continue
            if len(match.groups()) != 4:
                continue
            if not set(match.group(1)).issubset(ComputeTypes.size_family):
                continue
            if not set(match.group(3)).issubset(ComputeTypes.size_features):
                continue
            if not set(match.group(3)).issuperset(ComputeTypes.size_storage):
                continue
            if not set(match.group(4)).issubset(ComputeTypes.size_versions):
                continue
            if virtualization and not set(match.group(3)).issubset(ComputeTypes.vmp_features):
                continue
            if virtualization and not set(match.group(4)).issubset(ComputeTypes.vmp_versions):
                continue
            config_block = {'name': machine.name,
                            'cpu': machine.number_of_cores,
                            'memory': machine.memory_in_mb}
            machine_type_list.append(config_block)

        if len(machine_type_list) == 0:
            raise AzureDriverError(f"no machine types in location {location}")

        return machine_type_list

    def get_resources(self, location: str):
        result_list = self.compute_client.resource_skus.list()
        resource_list = [r for r in result_list if location in r.locations]
        return resource_list

    def get_machine_types(self, location: str, virtualization: bool = False):
        result_list = []
        machine_list = self.list(location, virtualization)
        machine_list = sorted(machine_list, key=lambda m: m['name'][-1], reverse=True)

        for machine_type in C.MACHINE_TYPES:
            machine = next((m for m in machine_list if m['cpu'] == machine_type['cpu'] and m['memory'] == machine_type['memory']), None)
            if not machine:
                continue
            machine.update(dict(machine_type=machine_type['name']))
            result_list.append(machine)

        return result_list

    def get_machine(self, name: str, location: str, virtualization: bool = False):
        machine_list = self.get_machine_types(location, virtualization)
        return next((m for m in machine_list if m['machine_type'] == name), None)

    def check_capacity(self, resource_list: List, machine_size: str, location: str):
        for resource in resource_list:
            if machine_size != resource.name:
                continue
            zone_list = next((i.zones for i in resource.location_info if i.location == location), [])
            if set(zone_list) != set(self.azure_availability_zones):
                return False
            restriction = next((r.reason_code for r in resource.restrictions if location in r.values), None)
            if restriction and restriction == "NotAvailableForSubscription":
                return False
            return True

    def details(self, machine_type: str) -> Union[dict, None]:
        try:
            sizes = self.compute_client.virtual_machine_sizes.list(self.azure_location)
        except Exception as err:
            raise AzureDriverError(f"error getting machine type {machine_type}: {err}")

        for group in list(sizes):
            if group.name == machine_type:
                return {'name': group.name,
                        'cpu': int(group.number_of_cores),
                        'memory': int(group.memory_in_mb),
                        'disk': int(group.resource_disk_size_in_mb)}
        return None



##
##

import os.path
import re
import logging
import time
from itertools import cycle, islice
from couchformation.aws.driver.image import Image
from couchformation.aws.driver.machine import MachineType
from couchformation.aws.driver.instance import Instance
from couchformation.aws.driver.base import CloudBase
from couchformation.aws.driver.constants import aws_storage_matrix, aws_arch_matrix, PlacementType
from couchformation.aws.driver.dns import DNS
from couchformation.aws.driver.nsg import SecurityGroup
from couchformation.aws.network import AWSNetwork
from couchformation.deployment import MetadataManager
from couchformation.config import get_state_file, get_state_dir, PortSettingSet, State
from couchformation.exception import FatalError
from couchformation.kvdb import KeyValueStore
from couchformation.util import FileManager, Synchronize, UUIDGen, parameter_to_dict, csv_dict_concat

logger = logging.getLogger('couchformation.aws.node')
logger.addHandler(logging.NullHandler())


class AWSNodeError(FatalError):
    pass


class AWSDeployment(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.project = parameters.get('project')
        self.build = parameters.get('build')
        self.region = parameters.get('region')
        self.zone = parameters.get('zone')
        self.auth_mode = parameters.get('auth_mode')
        self.profile = parameters.get('profile')
        self.ssh_key = parameters.get('ssh_key')
        self.os_id = parameters.get('os_id')
        self.os_version = parameters.get('os_version')
        self.os_arch = parameters.get('os_arch') if parameters.get('os_arch') else 'x86_64'
        self.feature = parameters.get('feature')
        self.cloud = parameters.get('cloud')
        self.group = parameters.get('group')
        self.number = parameters.get('number')
        self.machine_type = parameters.get('machine_type')
        self.ports = parameters.get('ports')
        self.tags = parameters.get('tags') if parameters.get('tags') else ''
        self.allow = parameters.get('allow') if parameters.get('allow') else "0.0.0.0/0"
        self.volume_size = parameters.get('volume_size') if parameters.get('volume_size') else "256"
        self.volume_iops = parameters.get('volume_iops') if parameters.get('volume_iops') \
            else next((aws_storage_matrix[s] for s in aws_storage_matrix if s >= int(self.volume_size)), "3000")
        self.services = parameters.get('services') if parameters.get('services') else "default"

        project_uid = MetadataManager(self.project).project_uid
        self.asset_prefix = f"cf-{project_uid}"
        self.node_name = f"{self.name}-node-{self.number:02d}"
        node_code = UUIDGen().text_hash(self.node_name)
        self.node_encoded = f"{self.asset_prefix}-{node_code}-node"

        self.tags = csv_dict_concat({"Project": self.project}, self.tags)
        self.tags = csv_dict_concat({"Service": self.name}, self.tags)

        filename = get_state_file(self.project, self.name)

        with Synchronize():
            try:
                state_dir = get_state_dir(self.project, self.name)
                if not os.path.exists(state_dir):
                    FileManager().make_dir(state_dir)
            except Exception as err:
                raise AWSNodeError(f"can not create state dir: {err}")

        document = self.node_name
        self.state = KeyValueStore(filename, document)

        CloudBase(self.parameters).test_session()

        self.aws_network = AWSNetwork(self.parameters)

    def check_state(self):
        if self.state.get('instance_id'):
            result = Instance(self.parameters).details(self.state['instance_id'])
            if result is None:
                logger.warning(f"Removing stale state entry for instance {self.state['instance_id']}")
                del self.state['instance_id']
        if self.state.get('node_security_group_id'):
            result = SecurityGroup(self.parameters).details(self.state['node_security_group_id'])
            if result is None:
                logger.warning(f"Removing stale state entry for security group {self.state['node_security_group_id']}")
                del self.state['node_security_group_id']

    def deploy(self):
        self.check_state()
        subnet_list = []
        nsg_list = []

        self.state['state'] = State.DEPLOYING.value

        if self.state.get('instance_id'):
            logger.info(f"Node {self.node_name} already exists")
            return self.state.as_dict

        ssh_key_name = self.aws_network.ssh_key_id
        sg_id = self.aws_network.security_group_id
        nsg_list.append(sg_id)

        for n, zone_state in enumerate(self.aws_network.zones):
            subnet_list.append(dict(
                subnet_id=zone_state[2],
                zone=zone_state[0],
                cidr=zone_state[1],
            ))

        if len(subnet_list) == 0:
            raise AWSNodeError(f"can not get subnet list, check project settings")

        if not self.zone:
            subnet_cycle = cycle(subnet_list)
            subnet = next(islice(subnet_cycle, self.number - 1, None))
        else:
            subnet = next((z for z in subnet_list if z['zone'] == self.zone), None)

        if not subnet:
            raise AWSNodeError(f"Can not determine availability zone (check project settings)")

        image = Image(self.parameters).list_standard(os_id=self.os_id, os_version=self.os_version, architecture=self.os_arch, feature=self.feature)
        if not image:
            raise AWSNodeError(f"can not find image for type {self.os_id} {self.os_version}")

        logger.info(f"Using image {image['name']} type {image['os_id']} version {image['os_version']}")

        self.state['service'] = self.name
        self.state['username'] = image['os_user']

        machine_type = self.machine_type
        volume_iops = int(self.volume_iops)
        volume_size = int(self.volume_size)
        services = self.services

        machine = MachineType(self.parameters).get_machine(self.machine_type, self.os_arch)
        if not machine:
            raise AWSNodeError(f"can not find machine for type {machine_type}")
        machine_name = machine['name']
        machine_ram = int(machine['memory'] / 1024)
        logger.info(f"Selecting machine type {machine_name}")

        placement = PlacementType(aws_arch_matrix[self.os_arch])

        if placement == PlacementType.HOST:
            if not self.state.get('host_id'):
                host_list = Instance(self.parameters).list_hosts(machine_name)
                host_id = next((h['id'] for h in host_list if h['capacity'] == machine['cpu']), None)
                if host_id:
                    logger.info(f"Using dedicated host {host_id}")
                else:
                    host_name = f"{self.project}-host"
                    logger.info(f"Allocating dedicated host for machine type {machine_name}")
                    host_id = Instance(self.parameters).allocate_host(host_name, subnet['zone'], machine_name)
                    logger.info(f"Allocated host {host_id}")
                self.state['host_id'] = host_id
            else:
                host_id = self.state.get('host_id')
        else:
            host_id = None

        if self.ports:
            port_sg_id = self.aws_network.create_node_group_sg(self.name, self.group, self.ports.split(','))
            logger.info(f"Assigning service group security group {port_sg_id}")
            nsg_list.append(port_sg_id)

        build_ports = PortSettingSet().create().get(self.build)
        if build_ports:
            build_sg_id = self.aws_network.create_build_sg(self.build)
            logger.info(f"Assigning build security group {build_sg_id}")
            nsg_list.append(build_sg_id)

        if image['os_id'] == 'windows':
            win_sg_id = self.aws_network.create_win_sg()
            logger.info(f"Assigning windows security group {win_sg_id}")
            nsg_list.append(win_sg_id)
            enable_winrm = True
        else:
            enable_winrm = False

        logger.info(f"Creating node {self.node_name}")
        instance_id = Instance(self.parameters).run(self.node_encoded,
                                                    image['name'],
                                                    ssh_key_name,
                                                    nsg_list,
                                                    subnet['subnet_id'],
                                                    subnet['zone'],
                                                    swap_size=machine_ram,
                                                    data_size=volume_size,
                                                    data_iops=volume_iops,
                                                    instance_type=machine_name,
                                                    placement=placement,
                                                    host_id=host_id,
                                                    enable_winrm=enable_winrm,
                                                    tags=parameter_to_dict(self.tags))

        self.state['instance_id'] = instance_id
        self.state['name'] = self.node_encoded
        self.state['services'] = services
        self.state['zone'] = subnet['zone']
        self.aws_network.add_service(self.node_name)

        while True:
            try:
                instance_details = Instance(self.parameters).details(instance_id)
                self.state['public_ip'] = instance_details['PublicIpAddress']
                self.state['private_ip'] = instance_details['PrivateIpAddress']
                break
            except KeyError:
                time.sleep(1)

        if self.aws_network.public_zone and self.aws_network.domain_name and not self.state.get('public_hostname'):
            host_name = f"{self.node_name}.{self.aws_network.domain_name}"
            DNS(self.parameters).add_record(self.aws_network.public_zone, host_name, [self.state['public_ip']])
            self.state['public_zone_id'] = self.aws_network.public_zone
            self.state['public_hostname'] = host_name

        if self.aws_network.private_zone and self.aws_network.domain_name and not self.state.get('private_hostname'):
            host_name = f"{self.node_name}.{self.aws_network.domain_name}"
            DNS(self.parameters).add_record(self.aws_network.private_zone, host_name, [self.state['private_ip']])
            self.state['private_zone_id'] = self.aws_network.private_zone
            self.state['private_hostname'] = host_name

        if image['os_id'] == 'windows':
            password = Instance(self.parameters).get_password(instance_id, self.ssh_key)
            self.state['host_password'] = password

        logger.info(f"Created instance {instance_id}")
        self.state['state'] = State.DEPLOYED.value
        return self.state.as_dict

    def destroy(self):
        self.state['state'] = State.DESTROYING.value
        if self.state.get('public_hostname'):
            domain_id = self.state['public_zone_id']
            name = self.state['public_hostname']
            ip = self.state['public_ip']
            DNS(self.parameters).delete_record(domain_id, name, [ip])
            logger.info(f"Deleted DNS record for {ip}")
        if self.state.get('private_hostname'):
            domain_id = self.state['private_zone_id']
            name = self.state['private_hostname']
            ip = self.state['private_ip']
            DNS(self.parameters).delete_record(domain_id, name, [ip])
            logger.info(f"Deleted DNS record for {ip}")
        if self.state.get('instance_id'):
            instance_id = self.state['instance_id']
            Instance(self.parameters).terminate(instance_id)
            logger.info(f"Removed instance {instance_id}")
        if self.state.get('host_id'):
            host_id = self.state.get('host_id')
            host = Instance(self.parameters).get_host_by_id(host_id)
            if host['age'] >= 24:
                Instance(self.parameters).release_host(host_id)
                del self.state['host_id']
                logger.info(f"Released host {host_id}")
            else:
                logger.warning(f"Can not remove dedicated host {host_id} age {host['age']} hrs is less than 24")
        if self.state.get('node_security_group_id'):
            sg_id = self.state.get('node_security_group_id')
            SecurityGroup(self.parameters).delete(sg_id)
            del self.state['node_security_group_id']
            logger.info(f"Removing security group {sg_id}")
        self.state.clear()
        self.aws_network.remove_service(self.node_name)
        self.state['state'] = State.IDLE.value

    def info(self):
        return self.state.as_dict

    @staticmethod
    def _calc_iops(value: str):
        num = int(value)
        iops = num * 10
        return str(3000 if iops < 3000 else 16000 if iops > 16000 else iops)

    @staticmethod
    def _name_check(value):
        p = re.compile(r"^[a-z]([-_a-z0-9]*[a-z0-9])?$")
        if p.match(value):
            return value
        else:
            raise AWSNodeError("names must only contain letters, numbers, dashes and underscores")



##
##

import re
import logging
import time
from itertools import cycle, islice
from couchformation.gcp.driver.base import CloudBase
from couchformation.gcp.driver.instance import Instance
from couchformation.gcp.driver.machine import MachineType
from couchformation.gcp.driver.disk import Disk
from couchformation.gcp.driver.image import Image
from couchformation.gcp.driver.dns import DNS
from couchformation.gcp.network import GCPNetwork
from couchformation.deployment import MetadataManager
from couchformation.config import get_state_file, get_state_dir, PortSettingSet, State
from couchformation.ssh import SSHUtil
from couchformation.exception import FatalError
from couchformation.kvdb import KeyValueStore
from couchformation.util import FileManager, Synchronize, UUIDGen

logger = logging.getLogger('couchformation.gcp.node')
logger.addHandler(logging.NullHandler())


class GCPNodeError(FatalError):
    pass


class GCPDeployment(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.project = parameters.get('project')
        self.build = parameters.get('build')
        self.region = parameters.get('region')
        self.zone = parameters.get('zone')
        self.auth_mode = parameters.get('auth_mode')
        self.profile = parameters.get('profile')
        self.ssh_key = parameters.get('ssh_key')
        self.os_id = parameters.get('os_id')
        self.os_version = parameters.get('os_version')
        self.feature = parameters.get('feature')
        self.cloud = parameters.get('cloud')
        self.group = parameters.get('group')
        self.number = parameters.get('number')
        self.machine_type = parameters.get('machine_type')
        self.ports = parameters.get('ports')
        self.volume_size = parameters.get('volume_size') if parameters.get('volume_size') else "256"
        self.services = parameters.get('services') if parameters.get('services') else "default"

        project_uid = MetadataManager(self.project).project_uid
        self.asset_prefix = f"cf-{project_uid}"
        self.node_name = f"{self.name}-node-{self.number:02d}"
        self.swap_disk = f"{self.name}-swap-{self.number:02d}"
        self.data_disk = f"{self.name}-data-{self.number:02d}"
        node_code = UUIDGen().text_hash(self.node_name)
        self.node_encoded = f"{self.asset_prefix}-{node_code}-node"
        self.swap_encoded = f"{self.asset_prefix}-{node_code}-swap"
        self.data_encoded = f"{self.asset_prefix}-{node_code}-data"

        filename = get_state_file(self.project, self.name)

        with Synchronize():
            try:
                state_dir = get_state_dir(self.project, self.name)
                FileManager().make_dir(state_dir)
            except Exception as err:
                raise GCPNodeError(f"can not create state dir: {err}")

        document = self.node_name
        self.state = KeyValueStore(filename, document)

        self.gcp_network = GCPNetwork(self.parameters)
        self.gcp_base = CloudBase(self.parameters)

        self.gcp_project = self.gcp_base.gcp_project
        self.service_account_email = self.gcp_base.service_account_email
        self.account_email = self.gcp_base.account_email

    def check_state(self):
        if self.state.get('instance_id'):
            result = Instance(self.parameters).find(self.state['instance_id'])
            if result is None:
                logger.warning(f"Removing stale state entry for instance {self.state['instance_id']}")
                del self.state['instance_id']
        if self.state.get('swap_disk'):
            result = Disk(self.parameters).find(self.state['swap_disk'])
            if result is None:
                logger.warning(f"Removing stale state entry for disk {self.state['swap_disk']}")
                del self.state['swap_disk']
        if self.state.get('data_disk'):
            result = Disk(self.parameters).find(self.state['data_disk'])
            if result is None:
                logger.warning(f"Removing stale state entry for disk {self.state['data_disk']}")
                del self.state['data_disk']

    def deploy(self):
        self.check_state()
        subnet_list = []

        self.state['state'] = State.DEPLOYING.value

        if self.state.get('instance_id'):
            logger.info(f"Node {self.node_name} already exists")
            return self.state.as_dict

        ssh_pub_key_text = SSHUtil().get_ssh_public_key(self.ssh_key)
        vpc_name = self.gcp_network.network
        subnet_name = self.gcp_network.subnet

        for n, zone_state in enumerate(self.gcp_network.zones):
            subnet_list.append(dict(
                subnet_id=zone_state[1],
                zone=zone_state[0],
            ))

        if len(subnet_list) == 0:
            raise GCPNodeError(f"can not get subnet list, check project settings")

        if not self.zone:
            subnet_cycle = cycle(subnet_list)
            subnet = next(islice(subnet_cycle, self.number - 1, None))
        else:
            subnet = next((z for z in subnet_list if z['zone'] == self.zone), None)

        if not subnet:
            raise GCPNodeError(f"Can not determine availability zone (check project settings)")

        image = Image(self.parameters).list_standard(os_id=self.os_id, os_version=self.os_version)
        if not image:
            raise GCPNodeError(f"can not find image for type {self.os_id} {self.os_version}")

        logger.info(f"Using image {image['name']} type {image['os_id']} version {image['os_version']}")

        self.state['service'] = self.name
        self.state['username'] = image['os_user']

        machine_type = self.machine_type
        volume_size = self.volume_size
        services = self.services

        machine = MachineType(self.parameters).get_machine(self.machine_type, subnet['zone'])
        if not machine:
            raise GCPNodeError(f"can not find machine for type {machine_type}")
        machine_name = machine['name']
        machine_ram = str(machine['memory'] / 1024)
        logger.info(f"Selecting machine type {machine_name}")

        if self.feature == "vmp":
            logger.info(f"Enabling nested virtualization")
            virtualization = True
        else:
            virtualization = False

        if self.ports:
            self.gcp_network.create_node_group_sg(self.name, self.group, self.ports.split(','))
            logger.info("Requesting service group firewall rule")

        build_ports = PortSettingSet().create().get(self.build)
        if build_ports:
            self.gcp_network.create_build_sg(self.build)
            logger.info(f"Requesting build {self.build} firewall rule")

        if image['os_id'] == 'windows':
            self.gcp_network.create_win_sg()
            logger.info("Requesting windows firewall rule")

        logger.info(f"Creating disk {self.swap_encoded} ({self.swap_disk})")
        Disk(self.parameters).create(self.swap_encoded, subnet['zone'], machine_ram)
        self.state['swap_disk'] = self.swap_encoded

        logger.info(f"Creating disk {self.data_encoded} ({self.data_disk})")
        Disk(self.parameters).create(self.data_encoded, subnet['zone'], volume_size)
        self.state['data_disk'] = self.data_encoded

        logger.info(f"Creating node {self.node_encoded} ({self.node_name})")
        Instance(self.parameters).run(self.node_encoded,
                                      image['image_project'],
                                      image['name'],
                                      self.service_account_email,
                                      subnet['zone'],
                                      vpc_name,
                                      subnet_name,
                                      image['os_user'],
                                      ssh_pub_key_text,
                                      self.swap_encoded,
                                      self.data_encoded,
                                      machine_type=machine_name,
                                      virtualization=virtualization)

        self.state['instance_id'] = self.node_encoded
        self.state['name'] = self.node_encoded
        self.state['services'] = services
        self.state['zone'] = subnet['zone']
        self.gcp_network.add_service(self.node_name)

        while True:
            try:
                instance_details = Instance(self.parameters).details(self.node_encoded, subnet['zone'])
                self.state['public_ip'] = instance_details['networkInterfaces'][0]['accessConfigs'][0]['natIP']
                self.state['private_ip'] = instance_details['networkInterfaces'][0]['networkIP']
                break
            except KeyError:
                time.sleep(1)
            except TypeError:
                raise GCPNodeError(f"Failed to properly start node {self.node_encoded} - try removing and recreating service")

        if self.gcp_network.public_zone and self.gcp_network.domain_name and not self.state.get('public_hostname'):
            host_name = f"{self.node_name}.{self.gcp_network.domain_name}"
            DNS(self.parameters).add_record(self.gcp_network.public_zone, host_name, [self.state['public_ip']])
            self.state['public_zone_id'] = self.gcp_network.public_zone
            self.state['public_hostname'] = host_name

        if self.gcp_network.private_zone and self.gcp_network.domain_name and not self.state.get('private_hostname'):
            host_name = f"{self.node_name}.{self.gcp_network.domain_name}"
            DNS(self.parameters).add_record(self.gcp_network.private_zone, host_name, [self.state['private_ip']])
            self.state['private_zone_id'] = self.gcp_network.private_zone
            self.state['private_hostname'] = host_name

        if image['os_id'] == 'windows':
            password = Instance(self.parameters).gen_password(image['os_user'],
                                                              self.node_encoded,
                                                              subnet['zone'],
                                                              self.account_email,
                                                              self.ssh_key)
            self.state['host_password'] = password

        logger.info(f"Created instance {self.node_encoded}")
        self.state['state'] = State.DEPLOYED.value
        return self.state.as_dict

    def destroy(self):
        self.state['state'] = State.DESTROYING.value
        if self.state.get('public_hostname'):
            domain_id = self.state['public_zone_id']
            name = self.state['public_hostname']
            ip = self.state['public_ip']
            DNS(self.parameters).delete_record(domain_id, name)
            logger.info(f"Deleted DNS record for {ip}")
        if self.state.get('private_hostname'):
            domain_id = self.state['private_zone_id']
            name = self.state['private_hostname']
            ip = self.state['private_ip']
            DNS(self.parameters).delete_record(domain_id, name)
            logger.info(f"Deleted DNS record for {ip}")
        if self.state.get('instance_id'):
            instance_name = self.state['instance_id']
            zone = self.state['zone']
            Instance(self.parameters).terminate(instance_name, zone)
            logger.info(f"Removed instance {instance_name}")
            Disk(self.parameters).delete(self.state['swap_disk'], zone)
            logger.info(f"Removed disk {self.state['swap_disk']}")
            Disk(self.parameters).delete(self.state['data_disk'], zone)
            logger.info(f"Removed disk {self.state['data_disk']}")
            self.state.clear()
            self.gcp_network.remove_service(self.node_name)
            logger.info(f"Removed instance {instance_name}")
            self.state['state'] = State.IDLE.value

    def info(self):
        return self.state.as_dict

    @staticmethod
    def _name_check(value):
        p = re.compile(r"^[a-z]([-_a-z0-9]*[a-z0-9])?$")
        if p.match(value):
            return value
        else:
            raise GCPNodeError("names must only contain letters, numbers, dashes and underscores")



##
##

import socket
import ipaddress
from ipaddress import IPv4Network
import logging
from typing import Union


class NetworkDriver(object):

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.ip_space = []
        self.active_network: IPv4Network = ipaddress.ip_network("10.1.0.0/16")
        self.super_net: IPv4Network = ipaddress.ip_network("10.0.0.0/8")

    def set_active_network(self, cidr: str):
        self.active_network: IPv4Network = ipaddress.ip_network(cidr)

    def add_network(self, cidr: str) -> None:
        cidr_net = ipaddress.ip_network(cidr)
        self.ip_space.append(cidr_net)

    def get_next_subnet(self, prefix=24) -> str:
        for subnet in self.active_network.subnets(new_prefix=prefix):
            yield subnet.exploded

    def get_next_network(self) -> Union[str, None]:
        candidates = list(self.super_net.subnets(new_prefix=16))

        for network in self.ip_space:
            available = []
            for n, candidate in enumerate(candidates):
                try:
                    if network.prefixlen < 16:
                        result = list(network.address_exclude(candidate))
                    else:
                        result = list(candidate.address_exclude(network))
                except ValueError:
                    available.append(candidate)
            candidates = available

        if len(candidates) == 0:
            return None

        self.active_network = candidates[0]
        self.ip_space.append(self.active_network)
        return self.active_network.exploded


class NetworkUtil(object):

    @staticmethod
    def local_ip_address():
        try:
            socket.setdefaulttimeout(2)
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            local_ip = s.getsockname()[0]
            s.close()
            return local_ip
        except TimeoutError:
            return None



##
##

import logging
from itertools import cycle
from couchformation.network import NetworkDriver
from couchformation.config import get_state_file, get_state_dir, State
from couchformation.exception import FatalError
from couchformation.util import FileManager, synchronize
from couchformation.kvdb import KeyValueStore
from couchformation.docker.driver.network import Network

logger = logging.getLogger('couchformation.docker.network')
logger.addHandler(logging.NullHandler())
logging.getLogger("docker").setLevel(logging.WARNING)


class DockerNetworkError(FatalError):
    pass


class DockerNetwork(object):

    def __init__(self, parameters: dict):
        self.parameters = parameters
        self.name = parameters.get('name')
        self.project = parameters.get('project')
        self.cloud = parameters.get('cloud')

        filename = get_state_file(self.project, 'common')

        try:
            state_dir = get_state_dir(self.project, 'common')
            FileManager().make_dir(state_dir)
        except Exception as err:
            raise DockerNetworkError(f"can not create state dir: {err}")

        document = f"network:{self.cloud}"
        self.state = KeyValueStore(filename, document)

        self.docker_network = Network(self.parameters)

    def check_state(self):
        if self.state.get('network'):
            network = Network(self.parameters).get_network(self.state.get('network'))
            if network is None:
                logger.warning(f"Removing stale state entry for network {self.state.get('network')}")
                del self.state['network']

    @synchronize()
    def create_vpc(self):
        self.check_state()
        cidr_util = NetworkDriver()
        net_name = f"{self.project}-net"

        for net in self.docker_network.cidr_list:
            cidr_util.add_network(net)

        try:
            self.state['state'] = State.DEPLOYING.value

            if not self.state.get('network'):
                cidr_util.get_next_network()
                subnet_list = list(cidr_util.get_next_subnet())
                subnet_cycle = cycle(subnet_list[1:])
                subnet_cidr = next(subnet_cycle)
                Network(self.parameters).create(net_name, subnet_cidr)
                self.state['network'] = net_name
                self.state['network_cidr'] = subnet_cidr
                logger.info(f"Created network {net_name}")

            self.state['state'] = State.DEPLOYED.value
        except Exception as err:
            raise DockerNetworkError(f"Error creating network: {err}")

    @synchronize()
    def peer_vpc(self):
        logger.warning(f"Peering not implemented for cloud {self.cloud}")

    @synchronize()
    def unpeer_vpc(self):
        pass

    @synchronize()
    def destroy_vpc(self):
        if self.state.list_len('services') > 0:
            logger.info(f"Active services, leaving project network in place")
            return

        try:
            self.state['state'] = State.DESTROYING.value
            if self.state.get('network'):
                net_name = self.state.get('network')
                Network(self.parameters).delete(net_name)
                del self.state['network']
                del self.state['network_cidr']
                logger.info(f"Removed network {net_name}")

            self.state['state'] = State.IDLE.value
        except Exception as err:
            raise DockerNetworkError(f"Error removing network: {err}")

    def create(self):
        logger.info(f"Creating docker network for {self.project}")
        self.create_vpc()

    def destroy(self):
        logger.info(f"Removing docker network for {self.project}")
        self.destroy_vpc()

    def get(self, key):
        return self.state.get(key)

    @property
    def network(self):
        return self.state.get('network')

    def add_service(self, name):
        self.state.list_add('services', name)

    def remove_service(self, name):
        self.state.list_remove('services', name)



##
##

import logging
import json
from azure.core.exceptions import ResourceNotFoundError
from couchformation.azure.driver.base import CloudBase, AzureDriverError
from couchformation.azure.driver.constants import AzureImagePublishers

logger = logging.getLogger('couchformation.azure.driver.instance')
logger.addHandler(logging.NullHandler())
logging.getLogger("azure").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)

WIN_INIT_SCRIPT = """winrm quickconfig -q -force
winrm set winrm/config/service/auth '@{Basic="true"}'
$hostname = $env:computername
$certificateThumbprint = (New-SelfSignedCertificate -DnsName "${hostname}" -CertStoreLocation Cert:\LocalMachine\My).Thumbprint
winrm create winrm/config/Listener?Address=*+Transport=HTTPS "@{Hostname=`"${hostname}`"; CertificateThumbprint=`"${certificateThumbprint}`"}"
netsh advfirewall firewall add rule name="Windows Remote Management (HTTPS-In)" dir=in action=allow protocol=TCP localport=5986
"""


class Instance(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)

    def run(self,
            name: str,
            image_publisher: str,
            image_offer: str,
            image_sku: str,
            zone: str,
            nic_id: str,
            username: str,
            public_key: str,
            resource_group: str,
            root_disk_name: str,
            root_type="Premium_LRS",
            root_size=256,
            machine_type="Standard_D4_v3",
            password="Passw0rd!",
            ultra=False):
        if not resource_group:
            resource_group = self.azure_resource_group

        instance_info = self.details(name, resource_group)
        if instance_info:
            return instance_info

        image_os = next((o['os_id'] for o in AzureImagePublishers.publishers if o['name'] == image_publisher), None)

        image_block = {
            'publisher': image_publisher,
            'offer': image_offer,
            'sku': image_sku,
            'version': 'latest'
        }

        parameters = {
            'location': self.azure_location,
            'zones': [zone],
            'os_profile': {
                'computer_name': name,
                'admin_username': username
            },
            'hardware_profile': {
                'vm_size': machine_type
            },
            'storage_profile': {
                'image_reference': image_block,
                'os_disk': {
                    'name': root_disk_name,
                    'disk_size_gb': root_size,
                    'create_option': 'FromImage',
                    'managed_disk': {
                        'storage_account_type': root_type
                    }
                }
            },
            'network_profile': {
                'network_interfaces': [{
                    'id': nic_id,
                }]
            },
        }

        if image_os == 'windows':
            os_config_block = {
                'admin_password': password,
                'windows_configuration': {
                    'enable_automatic_updates': False
                }
            }
        else:
            os_config_block = {
                'linux_configuration': {
                    'ssh': {
                        'public_keys': [
                            {
                                'path': f"/home/{username}/.ssh/authorized_keys",
                                'key_data': public_key
                            }
                        ]
                    }
                }
            }

        parameters['os_profile'].update(os_config_block)
        if ultra:
            parameters['additional_capabilities'] = {}
            parameters['additional_capabilities']['ultra_ssd_enabled'] = True

        logger.debug(f"Creating instance {name} with parameters:\n{json.dumps(parameters, indent=2)}")
        try:
            request = self.compute_client.virtual_machines.begin_create_or_update(resource_group, name, parameters)
            request.wait()
            if image_os == 'windows':
                result = self.run_command(WIN_INIT_SCRIPT, name, resource_group)
                if result.exit_code != 0:
                    raise AzureDriverError(f"error running instance config script: {result.error}")
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"error creating instance: {err}")

    def attach_disk(self, instance: str, caching: str, lun: str, disk_id: str, resource_group: str):
        parameters = {
            'caching': caching,
            'lun': lun,
            'create_option': 'Attach',
            'managed_disk': {
                'id': disk_id
            }
        }

        try:
            vm = self.compute_client.virtual_machines.get(resource_group, instance)
            vm.storage_profile.data_disks.append(parameters)
            request = self.compute_client.virtual_machines.begin_create_or_update(resource_group, instance, vm)
            request.wait()
            return request.result()
        except Exception as err:
            raise AzureDriverError(f"error attaching disk: {err}")

    def details(self, instance: str, resource_group: str):
        try:
            machine = self.compute_client.virtual_machines.get(resource_group, instance)
            return machine
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error getting instance {instance}: {err}")

    def terminate(self, instance: str, resource_group: str) -> None:
        try:
            request = self.compute_client.virtual_machines.begin_delete(resource_group, instance)
            request.wait()
        except ResourceNotFoundError:
            return None
        except Exception as err:
            raise AzureDriverError(f"error deleting instance: {err}")

    def run_command(self, command: str, vm_name: str, resource_group: str):
        run_command_name = "RunPowerShellScript"

        run_command = {
            'command_id': 'RunPowerShellScript',
            'location': self.azure_location,
            'source': {
                'script': command
            }
        }

        request = self.compute_client.virtual_machine_run_commands.begin_create_or_update(resource_group, vm_name, run_command_name, run_command)
        request.wait()
        result = self.compute_client.virtual_machine_run_commands.get_by_virtual_machine(resource_group, vm_name, run_command_name, expand="instanceView")
        iw = result.instance_view
        return iw



##
##

import logging
import winrm
import socket
import time
import jinja2
from couchformation.exception import NonFatalError

logger = logging.getLogger('couchformation.provisioner.winrm')
logger.addHandler(logging.NullHandler())


class ProvisionerError(NonFatalError):
    pass


class WinRMProvisioner(object):

    def __init__(self, parameters: dict, command: str = '', root: bool = True):
        self.parameters = parameters
        self.command = command
        self.root = root
        self.service = self.parameters.get('service')
        self.project = self.parameters.get('project')
        self.username = self.parameters.get('username')
        self.public_ip = self.parameters.get('public_ip')
        self.private_ip = self.parameters.get('private_ip')
        self.public_hostname = self.parameters.get('public_hostname')
        self.private_hostname = self.parameters.get('private_hostname')
        self.public = self.parameters.get('public') if 'public' in self.parameters else False
        self.zone = self.parameters.get('zone')
        self.password = self.parameters.get('password') if 'password' in self.parameters else 'password'
        self.host_password = self.parameters.get('host_password')
        self.upload_file = self.parameters.get('upload')
        self.sw_version = self.parameters.get('sw_version') if 'sw_version' in self.parameters else 'latest'
        self.services = self.parameters.get('services')
        self.options = self.parameters.get('options') if self.parameters.get('options') else 'default'
        self.connect = ','.join(self.parameters.get('connect')) \
            if self.parameters.get('connect') and type(self.parameters.get('connect')) is list \
            else self.parameters.get('connect')
        self.private_ip_list = ','.join(self.parameters.get('private_ip_list'))
        self.public_ip_list = ','.join(self.parameters.get('public_ip_list'))
        self.service_list = ':'.join(self.parameters.get('service_list'))
        if self.parameters.get('private_host_list') and len(self.parameters.get('private_host_list')) > 0:
            self.private_host_list = ','.join(self.parameters.get('private_host_list'))
        else:
            self.private_host_list = 'null'
        if self.parameters.get('public_host_list') and len(self.parameters.get('public_host_list')) > 0:
            self.public_host_list = ','.join(self.parameters.get('public_host_list'))
        else:
            self.public_host_list = 'null'
        self.use_private_ip = self.parameters.get('use_private_ip') if self.parameters.get('use_private_ip') else False

        if not self.host_password:
            raise ProvisionerError('WinRM provisioner requires host_password property to be set')

    @staticmethod
    def upload():
        logger.warning("Upload not implemented")

    @staticmethod
    def copy_file(fl, target):
        logger.warning("File copy not implemented")

    def run(self):
        if self.use_private_ip:
            hostname = self.private_ip
        else:
            hostname = self.public_ip

        url = f"https://{hostname}:5986/wsman"

        if not self.wait_port(hostname):
            raise ProvisionerError(f"Host {hostname} is not reachable")

        logger.info(f"Connection to {hostname} successful")
        time.sleep(0.5)

        _command = self.resolve_variables(self.command)

        logger.info(f"Connecting to {hostname} as {self.username}")
        logger.debug(f"Running command: {_command}")

        s = winrm.Session(url, auth=(self.username, self.host_password), transport='ntlm', server_cert_validation='ignore')

        if not self.wait_service(s):
            raise ProvisionerError(f"WinRM service unavailable on {hostname}")

        r = s.run_ps(_command)

        for line in r.std_out.decode('utf-8').splitlines():
            line_out = line.strip()
            if len(line_out) == 0:
                continue
            log_out = f"{hostname}: {line_out}"
            logger.info(log_out)

        return r.status_code

    @staticmethod
    def wait_port(address: str, port: int = 5986, retry_count=300, factor=0.1):
        for retry_number in range(retry_count + 1):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(1)
            result = sock.connect_ex((address, port))
            sock.close()
            if result == 0:
                return True
            else:
                if retry_number == retry_count:
                    return False
                logger.info(f"Waiting for {address} to become reachable")
                wait = factor
                wait *= (2 ** (retry_number + 1))
                time.sleep(wait)

    @staticmethod
    def wait_service(session: winrm.Session, retry_count=300, factor=0.1):
        for retry_number in range(retry_count + 1):
            try:
                session.run_cmd('whoami')
                return True
            except Exception as err:
                logger.debug(f"WinRM unavailable: {err}")
                if retry_number == retry_count:
                    return False
                logger.info(f"Waiting for provisioner service")
                wait = factor
                wait *= (2 ** (retry_number + 1))
                time.sleep(wait)

    def resolve_variables(self, line: str):
        env = jinja2.Environment(undefined=jinja2.DebugUndefined)
        raw_template = env.from_string(line)
        formatted_value = raw_template.render(
            SERVICE_NAME=self.service,
            SOFTWARE_VERSION=self.sw_version,
            PASSWORD=self.password,
            PRIVATE_IP_LIST=self.private_ip_list,
            PUBLIC_IP_LIST=self.public_ip_list,
            SERVICE_LIST=self.service_list,
            OPTIONS=self.options,
            IP_LIST=self.public_ip_list if self.public else self.private_ip_list,
            HOST_LIST=self.public_host_list if self.public else self.private_host_list,
            NODE_ZONE=self.zone,
            SERVICES=self.services,
            CONNECT_SERVICE=self.connect,
            CONNECT_IP=self.connect,
            CONNECT_LIST=self.connect if self.connect else '127.0.0.1'
        )
        return formatted_value



##
##

import logging
import botocore.exceptions
from typing import Union, List
from couchformation.aws.driver.base import CloudBase, AWSDriverError, EmptyResultSet
from couchformation.aws.driver.constants import AWSTagStruct, AWSTag

logger = logging.getLogger('couchformation.aws.driver.gateway')
logger.addHandler(logging.NullHandler())
logging.getLogger("botocore").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.ERROR)


class InternetGateway(CloudBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def list(self) -> List[dict]:
        ig_list = []
        igs = []
        extra_args = {}

        try:
            while True:
                result = self.ec2_client.describe_internet_gateways(**extra_args)
                igs.extend(result['InternetGateways'])
                if 'NextToken' not in result:
                    break
                extra_args['NextToken'] = result['NextToken']
        except Exception as err:
            raise AWSDriverError(f"error getting Internet Gateway list: {err}")

        for ig_entry in igs:
            ig_block = {'owner': ig_entry['OwnerId'],
                        'attachments': [a['VpcId'] for a in ig_entry['Attachments']],
                        'id': ig_entry['InternetGatewayId']}
            ig_list.append(ig_block)

        if len(ig_list) == 0:
            raise EmptyResultSet(f"no Internet Gateways found")
        else:
            return ig_list

    def create(self, name: str, vpc_id: str, tags: dict = None) -> str:
        ig_tag = AWSTagStruct.build("internet-gateway")
        ig_tag.add(AWSTag("Name", name))
        if tags:
            for k, v in tags.items():
                ig_tag.add(AWSTag(k, str(v)))
        try:
            result = self.ec2_client.create_internet_gateway(TagSpecifications=[ig_tag.as_dict])
            ig_id = result['InternetGateway']['InternetGatewayId']
            self.ec2_client.attach_internet_gateway(InternetGatewayId=ig_id, VpcId=vpc_id)
        except Exception as err:
            raise AWSDriverError(f"error creating Internet Gateway: {err}")

        return ig_id

    def delete(self, ig_id: str) -> None:
        try:
            ig_data = self.details(ig_id)
            if not ig_data:
                return
            for vpc_id in ig_data.get('attachments', []):
                self.ec2_client.detach_internet_gateway(InternetGatewayId=ig_id, VpcId=vpc_id)
            self.ec2_client.delete_internet_gateway(InternetGatewayId=ig_id)
        except Exception as err:
            raise AWSDriverError(f"error deleting Internet Gateway: {err}")

    def get(self, name: str):
        get_filter = {
            'Name': "tag:Name",
            'Values': [
                name,
            ]
        }
        try:
            result = self.ec2_client.describe_internet_gateways(Filters=[get_filter])
            return result.get('InternetGateways', [])[0]['InternetGatewayId']
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting Internet Gateway details: {err}")

    def details(self, ig_id: str) -> Union[dict, None]:
        try:
            result = self.ec2_client.describe_internet_gateways(InternetGatewayIds=[ig_id])
            ig_entry = result['InternetGateways'][0]
            ig_block = {'owner': ig_entry['OwnerId'],
                        'attachments': [a['VpcId'] for a in ig_entry['Attachments']],
                        'id': ig_entry['InternetGatewayId']}
            return ig_block
        except IndexError:
            return None
        except botocore.exceptions.ClientError as err:
            if err.response['Error']['Code'].endswith('NotFound'):
                return None
            raise AWSDriverError(f"ClientError: {err}")
        except Exception as err:
            raise AWSDriverError(f"error getting Internet Gateway details: {err}")
